#
# Copyright (c) 2013 by DarkDefender <darkdefende (a) gmail.com>
# File bugs here: https://github.com/DarkDefender/weetweet

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import sys
import ast
import re
import os
import tempfile
import time
import calendar
import socket
import getpass
import string
import itertools

# This twitter plugin can be extended even more. Just look at the twitter api
# doc here: https://dev.twitter.com/docs/api/1.1

# Thanks to ttytter for interface ideas!
# http://www.floodgap.com/software/ttytter/

# I've borrowed some ideas, functions for ainmosni and his twitter plugin
# https://github.com/ainmosni/weetwit

weechat_call = True
import_ok = True
required_twitter_version = "1.14.1"

try:
    import weechat
    #We need to import sys to see which python version weechat runs (2 or 3)
    import sys
    is_py3 = sys.version_info >= (3, 0)
except:
    #import html parser so we can convert html strings to plain text
    try:
        import html.parser
    except:
        print("You need to have python3 installed to run this script!")
        exit(0)
    weechat_call = False

try:
    #Import python twitter lib
    from twitter import *
    from twitter.stream import Timeout, HeartbeatTimeout, Hangup
except:
    import_ok = False

try:
    #Import for version checking
    from pkg_resources import parse_version, get_distribution
    version = get_distribution("twitter").version
    if parse_version(required_twitter_version) > parse_version(version):
        import_ok = False
except:
    if weechat_call:
        weechat.prnt("", "You need to have pkg_resources installed for version checking")
    else:
        print("You need to have pkg_resources installed for version checking")

# These two keys is what identifies this twitter client as "weechat twitter"
# If you want to change it you can register your own keys at:
# https://dev.twitter.com/apps/new

CONSUMER_SECRET = 'ivx3oxxkSOAOofRuhmGXQK4nkLFNXD94wbJiRUBhN1g'
CONSUMER_KEY = 'NVkYe8DAeaw6YRcjw662ZQ'

script_options = {
    "oauth_token" : ("", "your twitter oauth token"),
    "oauth_secret" : ("", "your twitter oath secret"),
    "auth_complete" : (False, "whether you've finished the auth process or not"),
    "screen_name" : ("", "your twitter username"),
    "last_id" : ("", "the ID of the most recent tweet"),
    "print_id" : (True, "display short tweet indexes"),
    "alt_rt_style" : (False, "use alternate retweet display style"),
    "home_replies" : (False, "show replies on the home timeline"),
    "tweet_nicks" : (True, "add twitter usernames to the nicklist"),
    "index_pattern": ("an", "a pattern of character groups to build "
                            "short tweet indexes from. the default 'an' generates "
                            "looping indexes from a0 to z9. a = [a-z], n = [0-9]"),
}

#TODO have a dict for each buffer
tweet_dict = {'cur_index': None}
#Mega command dict
command_dict = dict(
    user="u", replies="r", view_tweet="v", thread="th", link="l",
    retweet="rt", delete="d", tweet="t", reply="re", new_tweets="new",
    follow_user="follow", unfollow_user="unfollow", following="f",
    followers="fo", about="a", block="b", unblock="ub",
    blocked_users="blocks", favorite="fav", unfavorite="unfav",
    favorites="favs", fav_retweet="fart", rate_limits="limits",
    home_timeline="home", clear_nicks="cnicks", clear_buffer="clear",
    create_stream="stream", restart_home_stream="re_home")
desc_dict = dict(
    user="<user> [<id> <count>|<id>|<count>], Request user timeline, " +
    "if <id> is given it will get tweets older than <id>, " +
    "<count> is how many tweets to get, valid number is 1-200",
    replies="[<id> <count>|<id>|<count>], Get any replies/mentions of you " +
    "if <id> is given it will get tweets older than <id>, " +
    "<count> is how many tweets to get, valid number is 1-200",
    view_tweet="<id>, View/get tweet with <id>",
    thread="<id>, View/get the reply chain of tweets (the thread) where " +
    "<id> is the last tweet in the thread.",
    link="<id>, Get a link to tweet with <id>",
    retweet="<id>, Retweet <id>",
    delete="<id>, Delete tweet <id>. You can only delete your own tweets...",
    tweet="<text>Tweet the text following this command",
    reply="<id> <text>, reply to <id>. You need to have @<username> " +
    "of the user that you reply to in the tweet text. If this is not " +
    "the case this will be treated like a normal tweet instead.",
    new_tweets="Get new tweets from your home_timeline. This is only " +
    "useful if you have disconnected from the home twitter stream",
    follow_user="<user>, Add user to people you follow",
    unfollow_user="<user>, Remove user for people you follow",
    following="[|<id>|<user>|<user> <id>], Show 'friends' of <user> or " +
    "if no user were given show the people you follow. If not all " +
    "followers were printed supply the <id> of the last list to get " +
    "the new batch of nicks",
    followers="[|<id>|<user>|<user> <id>], Who followes <user> or " +
    "if no user were given show your follower. If not all " +
    "followers were printed supply the <id> of the last list to get " +
    "the new batch of nicks",
    about="<user>, Print info about <user>",
    block="<user>, Block <user>",
    unblock="<user>, Unblock <user>",
    blocked_users="Print a list of users you have currently blocked",
    favorite="<id>, Add tweet <id> to you favorites",
    unfavorite="<id>, Remove tweet <id> from your favorites",
    favorites="[|<user>][<id> <count>|<id>|<count>], Request <user> favs, " +
    "if <user> is not given get your own favs. " +
    "If <id> is given it will get tweets older than <id>, " +
    "<count> is how many tweets to get, valid number is 1-200",
    fav_retweet="<id>, favorites and retweets <id>",
    rate_limits="[|<sub_group>], get the current status of the twitter " +
    "api limits. It prints how much you have left/used. " +
    " if <sub_group> is supplied it will only get/print that sub_group.",
    home_timeline="[<id> <count>|<id>|<count>], Get tweets from you home " +
    "timeline" +
    "if <id> is given it will get tweets older than <id>, " +
    "<count> is how many tweets to get, valid number is 1-200",
    clear_nicks="Clear nicks from the 'Tweet_parse' nick group. "+
    "These nicks are parsed from recived tweets, it can get " +
    "messy at times...",
    clear_buffer="Clear the twitter buffer of text "+
    "same as '/buffer clear'",
    create_stream="Create a twitter stream with the following filter "+
    "options: <users to stream> <keywords>. Note that the user list and the "+
    "keyword list must be comma seperated. IE user1, user2, user3 keyword1, keyword2.\n" +
    "If you only want to stream keywords, write '&' as 'users' input.\n"+
    "NOTE: you can only have one stream at a time because "+
    "twitter will IP ban you if you repeatedly request more than one "+
    "stream.",
    restart_home_stream="Restart the home timeline stream after it has " +
    "shutdown.")

SCRIPT_NAME = "weetweet"
SCRIPT_FILE_PATH = os.path.abspath(__file__)

twit_buf = ""
sock_hooks = {}
proc_hooks = {}
sock_fd_dict = {}
tweet_nicks_group = {}
friends_nicks_group = {}

html_escape_table = {
    '"': "&quot;",
    "'": "&apos;",
    }

def html_escape(text):
    """Produce entities within text."""
    return "".join(html_escape_table.get(c, c) for c in text)

def index_gen():
    #Generates indexes starting from a0 going up to z9, then repeating
    #index_pattern is a pattern of character groups to use for each part of the index
    # a = [a-z]
    # n = [0-9]
    #This setting needs a script reload to take effect
    index_parts = []
    for p in script_options['index_pattern']:
        if p == 'a':
            index_parts.append(string.ascii_lowercase)
        elif p == 'n':
            index_parts.append(string.digits)

    while True:
        for index in itertools.product(*index_parts):
            yield "".join(index)
#Create our index generator
index_generator = index_gen()

def dict_tweet(tweet_id):
    cur_index = tweet_dict['cur_index']
    if not tweet_id in tweet_dict.values():
        cur_index = next(index_generator)
        tweet_dict[cur_index] = tweet_id
        tweet_dict['cur_index'] = cur_index
        return cur_index
    else:
        for index, t_id in tweet_dict.items():
            if t_id == tweet_id:
                return index

    return None

def read_config():
    for item in script_options:
        script_options[item] = weechat.config_string(
            weechat.config_get("plugins.var.python."+SCRIPT_NAME+"." + item))
    for item in ["auth_complete", "print_id", "alt_rt_style", "home_replies", "tweet_nicks"]:
        #Convert to bool
        script_options[item] = weechat.config_string_to_boolean(script_options[item])
    for item in ["oauth_token", "oauth_secret"]:
        #Convert potentially encrypted tokens
        script_options[item] = weechat.string_eval_expression(script_options[item], {}, {}, {})

def config_cb(data, option, value):
    """Callback called when a script option is changed."""
    # for example, read all script options to script variables...
    # ...
    read_config()
    return weechat.WEECHAT_RC_OK

def add_to_nicklist(buf, nick, group=""):
    """Add nick to the nicklist."""
    if group == "":
        group = friends_nicks_group[buf]
    weechat.nicklist_add_nick(buf, group, nick, 'bar_fg', '', '', 1)

def remove_from_nicklist(buf, nick, group=""):
    """Remove nick from the nicklist."""
    if group == "":
        group = friends_nicks_group[buf]
    nick_ptr = weechat.nicklist_search_nick(buf, group, nick)
    weechat.nicklist_remove_nick(buf, nick_ptr)

def parse_for_nicks(text, buffer):
    #Parse text for twitter nicks and add them to nicklist
    regex = re.compile(r'@([A-Za-z0-9_]+)')
    for nick in regex.findall(text):
        add_to_nicklist(buffer, nick, tweet_nicks_group[buffer])

def print_tweet_data(buffer, tweets, data):

    for message in tweets:
        nick = message[1]
        text = message[3]
        reply_id = ""
        if script_options['tweet_nicks']:
            parse_for_nicks(text, buffer)
            add_to_nicklist(buffer, nick, tweet_nicks_group[buffer])

        if script_options['print_id']:
            t_id = weechat.color('reset') + ' ' + dict_tweet(message[2])
        else:
            t_id = ''

        if len(message) == 5:
            #This is a reply to a tweet
            arrow_col = weechat.color('chat_prefix_suffix')
            reset_col = weechat.color('reset')
            reply_id = arrow_col + "<" + reset_col + dict_tweet(
                message[4]) + arrow_col + "> " + reset_col
            temp_text = text
            text = reply_id
            reply_id = temp_text

        weechat.prnt_date_tags(buffer, message[0], "notify_message",
                               "%s%s\t%s%s" % (nick, t_id, text, reply_id))
    if data == "id":
        try:
            if( script_options['last_id'] == "" or
                int(script_options['last_id']) < int(tweets[-1][2]) ):

                script_options['last_id'] = tweets[-1][2]
                # Save last id
                weechat.config_set_plugin("last_id", script_options["last_id"])
        except:
            pass

def trim_tweet_data(tweet_data, screen_name, alt_rt_style, is_py3):
    # Because of the huge amount of data, we need to cut down on most of it
    # because we only really want a small subset of it. This also prevents the
    # output buffer from overflowing when fetching many tweets at once.
    h = html.parser.HTMLParser()

    output = []
    for message in tweet_data:
        if message.get('retweeted_status'):
            if alt_rt_style:
                if message['user']['screen_name'] == screen_name:
                    #escape highlighting
                    message['user']['screen_name'] = "<you>"
                if message.get('full_text'):
                    message['full_text'] = message['retweeted_status']['full_text'] + \
                                " (retweeted by " + message['user']['screen_name'] + ")"
                else:
                    message['text'] = message['retweeted_status']['text'] + \
                            " (retweeted by " + message['user']['screen_name'] + ")"
                message['user'] = message['retweeted_status']['user']
            else:
                if message.get('full_text'):
                    message['full_text'] = "RT @{}: {}".format(
                        message['retweeted_status']['user']['screen_name'],
                        message['retweeted_status']['full_text'])
                else:
                    message['text'] = "RT @{}: {}".format(
                        message['retweeted_status']['user']['screen_name'],
                        message['retweeted_status']['text'])
        mes_list = [calendar.timegm(time.strptime(message['created_at'],
                                                  '%a %b %d %H:%M:%S +0000 %Y')),
                    message['user']['screen_name'],
                    message['id_str']]
        if is_py3:
            if message.get('full_text'):
                mes_list.append(h.unescape(message['full_text']))
            else:
                mes_list.append(h.unescape(message['text']))
        else:
            #convert text to bytes so python2 can read it correctly
            if message.get('full_text'):
                mes_list.append(h.unescape(message['full_text']).encode('utf-8'))
            else:
                mes_list.append(h.unescape(message['text']).encode('utf-8'))
        if message["in_reply_to_status_id_str"] != None:
            mes_list.append(message["in_reply_to_status_id_str"])

        output.append(mes_list)

    output.reverse()
    return output

def stream_message(buffer, tweet):
    #TODO because we can not use userstreams anymore, this code will probably never be called
    if 'delete' in tweet:
        # Only show the 'Got request to delete' dialog if it didn't come from us
        #Colorize the tweet id
        arrow_col = weechat.color('chat_prefix_suffix')
        reset_col = weechat.color('reset')
        dict_id = dict_tweet(tweet['delete']['status']['id_str'])
        id_str = arrow_col +  "<" + reset_col + dict_id + arrow_col + "> " + reset_col
        weechat.prnt(buffer, "%s%s" % (weechat.prefix("network"),
                                       "Got request to delete: " + id_str))
    elif 'event' in tweet:
        event_str = tweet['event']

        if event_str[-1] == "e":
            event_str += "d"
        elif event_str[-1] != "d":
            event_str += "ed"

        extra_str = ""

        if 'target_object' in tweet:
            if 'id_str' in tweet['target_object']:
                arrow_col = weechat.color('chat_prefix_suffix')
                reset_col = weechat.color('reset')
                dict_id = dict_tweet(tweet['target_object']['id_str'])
                extra_str = "'s tweet " + arrow_col +  "<" + reset_col + dict_id + arrow_col + \
                        "> " + reset_col

        #TODO make the event printing better
        weechat.prnt_date_tags(
            buffer, 0, "no_highlight", "%s%s" % (weechat.prefix("network"),
                                                 tweet['source']['screen_name'] + \
                                                         " " + event_str + \
                                            " " + tweet['target']['screen_name'] + extra_str))
    elif 'friends' in tweet:
        #This should be the initital message you get listing
        #friend ids when connecting to the twitter stream
        weechat.prnt(buffer, "%s%s" % (weechat.prefix("network"),
                                       "Connected to twitter streaming API."))
        #Get new tweets since the last update
        #(we might have missed some if this is after a reconnect)
        #Get latest tweets from timeline
        my_command_cb("silent", buffer, "new")
    else:
        weechat.prnt(buffer, "%s%s" % (weechat.prefix("network"),
                                       "recv stream data: " + str(tweet)))

def twitter_stream_cb(buffer, fd):

    #accept connection
    server = sock_fd_dict[sock_fd_dict[str(fd)]]
    conn, addr = server.accept()
    tweet = b""
    data = True
    while data:
        try:
            data = conn.recv(1024)
            tweet += data
        except:
            break

    try:
        tweet = tweet.decode('utf-8')
        tweet = ast.literal_eval(tweet)
    except:
        weechat.prnt(buffer, "Error recv stream message:")
        weechat.prnt(buffer, tweet)
        return weechat.WEECHAT_RC_OK
    #Is this a text message (normal tweet)?
    if isinstance(tweet, list):
        if buffer == twit_buf:
            #Update last recv id
            print_tweet_data(buffer, tweet, "id")
        else:
            print_tweet_data(buffer, tweet, "")
    elif tweet == "options":
        #We need to send over the stream options

        options = dict(screen_name=script_options['screen_name'],
                       name=sock_fd_dict[str(fd)],
                       alt_rt_style=int(script_options['alt_rt_style']),
                       home_replies=int(script_options['home_replies']),
                       token=script_options["oauth_token"],
                       secret=script_options["oauth_secret"],
                       last_id=script_options['last_id'],
                       is_py3=is_py3)

        if is_py3:
            conn.sendall(bytes(str(options), "utf-8"))
        else:
            conn.sendall(bytes(str(options)))
    else:
        #https://dev.twitter.com/docs/streaming-apis/messages
        #TODO handle stream events
        stream_message(buffer, tweet)

    conn.close()
    return weechat.WEECHAT_RC_OK

def twitter_stream(cmd_args):
    if len(cmd_args) < 3:
        return "Invalid stream command"
    if not os.path.exists(cmd_args[2]):
        return "The socket file doesn't exist! " + cmd_args[2]

    def connect():
        client = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        client.connect(cmd_args[2])
        #Don't block by default, timeout if no data is present
        client.setblocking(False)
        return client

    client = connect()
    client.settimeout(3)
    client.sendall(bytes('"options"', "utf-8"))
    client.shutdown(socket.SHUT_WR)

    data = True
    options = b""

    while data:
        try:
            data = client.recv(1024)
            options += data
        except:
            break

    try:
        options = options.decode('utf-8')
        option_dict = ast.literal_eval(options)
    except:
        return "Did not manage to get startup arguments to stream"

    client.close()

    oauth_token = option_dict['token']
    oauth_secret = option_dict['secret']
    home_replies = option_dict['home_replies']
    alt_rt_style = option_dict['alt_rt_style']
    screen_name = option_dict['screen_name']
    name = option_dict['name']
    last_id = option_dict['last_id']
    is_py3 = option_dict['is_py3']

    if len(cmd_args) >= 4:
        stream_args = cmd_args[3:]
    else:
        stream_args = ""

    # These arguments are optional. But the current code only handles this
    # configuration. So it's defined here if the defaults change.
    stream_options = dict(timeout=None, block=True, heartbeat_timeout=90)

    # Reconnect timer list [1sec, 5sec, 10sec, 1min, 2min]
    re_timer = [1, 2, 10, 60, 120]
    re_timer_idx = 0

    kwargs = dict(tweet_mode='extended')

    if name == "twitter":
        #Twitter has removed home user streams, so we have to do this instead...
        twitter = Twitter(auth=OAuth(
            oauth_token, oauth_secret, CONSUMER_KEY, CONSUMER_SECRET),
                          api_version='1.1', domain='api.twitter.com',
                          retry=True)

        while True:
            #TODO check with limits for how often we can make requests and how many we have left.
            if last_id == "":
                tweet_data = twitter.statuses.home_timeline(exclude_replies=no_home_replies,
                                                            tweet_mode='extended')
            else:
                tweet_data = twitter.statuses.home_timeline(
                    since_id=last_id,
                    count=200,
                    exclude_replies=home_replies,
                    tweet_mode='extended')
            if not tweet_data == []:
                tweet = trim_tweet_data(tweet_data,
                                        screen_name,
                                        alt_rt_style,
                                        is_py3)
                client = connect()
                client.sendall(bytes(str(tweet), "utf-8"))
                client.close()
                stream_end_message = "Text message"

                #Update last_id (for the stream process)
                last_id = tweet[-1][2]

            #We have to wait at least a minute because of the twitter rate limits
            time.sleep(65)

    while True:
        try:
            stream = TwitterStream(auth=OAuth(
                oauth_token, oauth_secret, CONSUMER_KEY, CONSUMER_SECRET),
                                   **stream_options)

            twitter = Twitter(auth=OAuth(
                oauth_token, oauth_secret, CONSUMER_KEY, CONSUMER_SECRET),
                              api_version='1.1', domain='api.twitter.com')

            if stream_args[0] != "&":
                follow = stream_args[0]
                kwargs['screen_name'] = follow
                twitter_data = twitter.users.lookup(**kwargs)
                follow_ids = ""
                for user in twitter_data:
                    follow_ids += user['id_str'] + ", "
                follow_ids = follow_ids[:-1]
                if len(stream_args) == 2:
                    track = stream_args[1]
                    kwargs['track'] = track
                    kwargs['follow'] = follow_ids
                    tweet_iter = stream.statuses.filter(**kwargs)
                else:
                    kwargs['follow'] = follow_ids
                    tweet_iter = stream.statuses.filter(**kwargs)
            else:
                track = stream_args[1]
                kwargs['track'] = track
                tweet_iter = stream.statuses.filter(**kwargs)
        except:
            stream_end_message = "Connection problem (could not connect to twitter)"
            break

        stream_end_message = "Unknown reason"

        # Iterate over the stream.
        for tweet in tweet_iter:
            # You must test that your tweet has text. It might be a delete
            # or data message.
            if tweet is None:
                stream_end_message = "'None' reply"
            elif tweet is Timeout:
                stream_end_message = "Timeout"
            elif tweet is HeartbeatTimeout:
                stream_end_message = "Heartbeat Timeout"
            elif tweet is Hangup:
                stream_end_message = "Hangup"
            elif tweet.get('full_text') or tweet.get('text'):
                tweet = trim_tweet_data([tweet],
                                        screen_name,
                                        alt_rt_style,
                                        is_py3)
                client = connect()
                client.sendall(bytes(str(tweet), "utf-8"))
                client.close()
                stream_end_message = "Text message"
                # Reset the reconnect timer index when we get a new message
                re_timer_idx = 0
            else:
                #Got a other type of message
                client = connect()
                client.sendall(bytes(str(tweet), "utf-8"))
                client.close()
                stream_end_message = "Unhandled type message"

        client = connect()
        client.sendall(bytes(
            '"Disconnected, trying to reconnect in {0} sec. Reason: {1}"'.format(
                re_timer[re_timer_idx],
                stream_end_message),
            "utf-8"))
        client.close()

        time.sleep(re_timer[re_timer_idx])
        if re_timer_idx < (len(re_timer) - 1):
            re_timer_idx += 1

    return "Stream shut down after: " + stream_end_message + \
            ". You'll have to restart the stream manually. (:re_home, if home stream)"

def stream_close_cb(name, buffer):
    global sock_fd_dict
    global proc_hooks
    weechat.unhook(sock_hooks[name])
    #Is the process already unhooked?
    if proc_hooks[name]:
        weechat.unhook(proc_hooks[name])

    #remove fd key
    for key, value in sock_fd_dict.items():
        if value == name:
            sock_fd_dict.pop(key, None)
            break
    sock_fd_dict[name].close()
    proc_hooks.pop(name, None)
    sock_fd_dict.pop(name, None)
    return weechat.WEECHAT_RC_OK

def create_stream(name, args=""):
    global sock_fd_dict
    global proc_hooks
    global sock_hooks

    if proc_hooks.get(name):
        return "Stream has already been created, close it before trying to open a new one"

    #Check if buffer exists
    buffer = weechat.buffer_search("python", name)
    if buffer == "":
        buffer = weechat.buffer_new(name, "buffer_input_cb", name, "stream_close_cb", name)
        setup_buffer(buffer)

    if not sock_fd_dict.get(name):
        file_name = tempfile.gettempdir() + "/we_tw_" + getpass.getuser() + "_" + name
        if os.path.exists(file_name):
            os.remove(file_name)

        server = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        server.bind(file_name)
        #Don't block, timeout if no data is present
        server.setblocking(False)
        server.listen(1)
        file_fd = server.fileno()
        sock_fd_dict[str(file_fd)] = name
        sock_fd_dict[name] = server
        sock_hooks[name] = weechat.hook_fd(file_fd, 1, 0, 0, "twitter_stream_cb", buffer)

    proc_hooks[name] = weechat.hook_process("python3 " + SCRIPT_FILE_PATH + " " + \
                                            "stream " + file_name + " " + args, 0,
                                            "my_process_cb", str([buffer, "Stream"]))
    return "Started stream"

def my_process_cb(data, command, rc, out, err):

    if rc == weechat.WEECHAT_HOOK_PROCESS_ERROR:
        weechat.prnt("", "Error with command '%s'" %
                     command.replace(script_options["oauth_token"],
                                     "<token>").replace(script_options["oauth_secret"], "<secret>"))
        if err != "":
            weechat.prnt("", "stderr: %s" % err)

        return weechat.WEECHAT_RC_OK

    data = ast.literal_eval(data)
    buffer = data[0]
    end_mes = data[1]

    if out != "":
        if out[0] != "[" and out[0] != "{":
            #If message is just a string print it
            weechat.prnt(buffer, "%s%s" % (weechat.prefix("network"), out))
            process_output = []
        else:
            process_output = ast.literal_eval(out)

        #List message
        if len(end_mes) >= 1 and end_mes[0] == "L":
            if isinstance(process_output[-1], int):
                t_id = dict_tweet(str(process_output[-1])) + "\t"
                process_output = process_output[:-1]
                more = " ..."
            else:
                t_id = weechat.prefix("network")
                more = ""

            for nick in process_output:
                if end_mes == "LYFollowing":
                    end_mes = "LYou are following"
                    add_to_nicklist(buffer, nick)
                elif script_options['tweet_nicks']:
                    add_to_nicklist(buffer, nick, tweet_nicks_group[buffer])
            weechat.prnt_date_tags(buffer, 0, "no_highlight",
                                   "%s%s: %s%s" % (t_id, end_mes[1:], process_output, more))
            return weechat.WEECHAT_RC_OK

        if end_mes == "About":
            weechat.prnt(buffer, "Nick: %s | Name: %s | Protected: %s" %
                         (process_output['screen_name'],
                          process_output['name'],
                          process_output['protected']))
            weechat.prnt(buffer, "Description: %s" %
                         process_output['description'])
            weechat.prnt(buffer, "Location: %s | Time zone: %s" %
                         (process_output['location'],
                          process_output['time_zone']))
            weechat.prnt(buffer, "Created at: %s | Verified user: %s" %
                         (process_output['created_at'],
                          process_output['verified']))
            weechat.prnt(buffer, "Following: %s | Followers: %s | Favourites: %s | Tweets: %s" %
                         (process_output['friends_count'],
                          process_output['followers_count'],
                          process_output['favourites_count'],
                          process_output['statuses_count']))
            weechat.prnt(buffer, "Are you currently following this person: %s" %
                         (process_output['following']))
            return weechat.WEECHAT_RC_OK

        elif end_mes == "Stream":
            global proc_hooks
            #Clean up the stream hooks
            name = weechat.buffer_get_string(buffer, "name")
            #Set the hook to none because it has already been unhooked
            proc_hooks[name] = None
            stream_close_cb(name, buffer)
            #TODO restart stream correctly
            #create_stream(name)
            return weechat.WEECHAT_RC_OK

        print_tweet_data(buffer, process_output, end_mes)

        if end_mes != "id" and end_mes != "":
            weechat.prnt(buffer, "%s%s" % (weechat.prefix("network"), end_mes))

    if err != "":
        weechat.prnt("", "stderr: %s" % err)

    return weechat.WEECHAT_RC_OK

def get_twitter_data(cmd_args):
    # Read the oauth token and auth with the twitter api.
    # Return the requested tweets
    no_home_replies = True
    alt_rt_style = False
    is_py3 = False
    screen_name = ""

    h = html.parser.HTMLParser()

    try:
        if cmd_args[-1][0] == "[":
            option_list = ast.literal_eval(cmd_args[-1])
            cmd_args.pop(-1)
            if "home_replies" in option_list:
                no_home_replies = False
            if "alt_rt_style" in option_list:
                alt_rt_style = True
            if "is_py3" in option_list:
                is_py3 = True
            screen_name = option_list[0]
    except:
        pass

    if len(cmd_args) < 4:
        return "Invalid command"

    oauth_token = cmd_args[1]
    oauth_secret = cmd_args[2]

    try:
        if cmd_args[3] == "auth":
            twitter = Twitter(
                auth=OAuth(oauth_token, oauth_secret, CONSUMER_KEY, CONSUMER_SECRET),
                format='', api_version=None)
            if len(cmd_args) == 5:
                #pin
                return twitter.oauth.access_token(oauth_verifier=cmd_args[4])
            return twitter.oauth.request_token()

        twitter = Twitter(auth=OAuth(
            oauth_token, oauth_secret, CONSUMER_KEY, CONSUMER_SECRET),
                          api_version='1.1', domain='api.twitter.com')

        if cmd_args[3] == "settings":
            #this only gets called from within weechat
            return twitter.account.settings()['screen_name']
        elif cmd_args[3] == "u":
            kwargs = dict(count=20, screen_name=cmd_args[4], tweet_mode='extended')
            if len(cmd_args) == 7:
                kwargs['count'] = int(cmd_args[6])
                kwargs['max_id'] = cmd_args[5]
            elif len(cmd_args) == 6:
                if int(cmd_args[5]) <= 200:
                    kwargs['count'] = int(cmd_args[5])
                else:
                    kwargs['max_id'] = cmd_args[5]
            tweet_data = twitter.statuses.user_timeline(**kwargs)
        elif cmd_args[3] == "r":
            if len(cmd_args) == 6:
                kwargs = dict(count=int(cmd_args[5]), max_id=cmd_args[4], tweet_mode='extended')
                tweet_data = twitter.statuses.mentions_timeline(**kwargs)
            elif len(cmd_args) == 5:
                if int(cmd_args[4]) <= 200:
                    kwargs = dict(count=int(cmd_args[4]), tweet_mode='extended')
                    tweet_data = twitter.statuses.mentions_timeline(**kwargs)
                else:
                    kwargs = dict(max_id=cmd_args[4], tweet_mode='extended')
                    tweet_data = twitter.statuses.mentions_timeline(**kwargs)
            else:
                tweet_data = twitter.statuses.mentions_timeline(tweet_mode='extended')
        elif cmd_args[3] == "v":
            tweet_data = [twitter.statuses.show._(cmd_args[4])(tweet_mode='extended')]
        elif cmd_args[3] == "th":
            #only fetch up to 20 tweets in one go
            tweets_left = 0
            tweet_id = cmd_args[4]
            tweet_data = []
            while tweets_left < 20:
                tweets_left += 1
                #TODO handle deleted tweets
                temp_tweet = twitter.statuses.show._(tweet_id)(tweet_mode='extended')
                tweet_data.append(temp_tweet)
                if temp_tweet["in_reply_to_status_id_str"] != None:
                    tweet_id = temp_tweet["in_reply_to_status_id_str"]
                else:
                    break
        elif cmd_args[3] == "l":
            tweet_data = [twitter.statuses.show._(cmd_args[4])()]
            output = "Link for tweet: https://twitter.com/{}/status/{}".format(
                tweet_data[0]['user']['screen_name'],
                tweet_data[0]['id_str'])
            return output
        elif cmd_args[3] == "rt":
            tweet_data = [twitter.statuses.retweet._(cmd_args[4])()]
        elif cmd_args[3] == "d":
            #deletes tweet made by the user _(...) converts the id string to a call
            #returns the tweet that was deleted (not a list(dict) just a dict)
            #make it into a list so we don't have to write special cases for this
            tweet_data = [twitter.statuses.destroy._(cmd_args[4])()]
        elif cmd_args[3] == "t":
            #returns the tweet that was sent (not a list(dict) just a dict)
            #make it into a list so we don't have to write special cases for this
            tweet_data = [twitter.statuses.update(status=h.unescape(cmd_args[4]))]
        elif cmd_args[3] == "re":
            tweet_data = [twitter.statuses.update(status=h.unescape(cmd_args[5]),
                                                  in_reply_to_status_id=cmd_args[4])]
        elif cmd_args[3] == "new":
            tweet_data = twitter.statuses.home_timeline(
                since_id=cmd_args[4],
                count=200,
                exclude_replies=no_home_replies,
                tweet_mode='extended')
            if tweet_data == []:
                return "No new tweets available."
        elif cmd_args[3] == "follow":
            tweet_data = []
            twitter.friendships.create(screen_name=cmd_args[4])
        elif cmd_args[3] == "unfollow":
            tweet_data = []
            twitter.friendships.destroy(screen_name=cmd_args[4])
        elif cmd_args[3] == "f" or cmd_args[3] == "fo":
            if len(cmd_args) == 6:
                kwargs = dict(screen_name=cmd_args[4], stringify_ids=True,
                              cursor=int(cmd_args[5]), count=250)
            else:
                kwargs = dict(screen_name=cmd_args[4], stringify_ids=True, cursor=-1, count=250)
            if cmd_args[3] == "f":
                tweet_data = twitter.friends.ids(**kwargs)
            else:
                tweet_data = twitter.followers.ids(**kwargs)
            kwargs['cursor'] = tweet_data['next_cursor']
            friend_ids = tweet_data['ids']
            friend_list = list()

            while len(friend_ids) > 100:
                tweet_data = twitter.users.lookup(user_id=", ".join(friend_ids[:100]))
                friend_ids = friend_ids[100:]
                for user in tweet_data:
                    friend_list.append(user['screen_name'])
            tweet_data = twitter.users.lookup(user_id=", ".join(friend_ids))
            for user in tweet_data:
                friend_list.append(user['screen_name'])

            if kwargs['cursor'] != 0:
                friend_list.append(kwargs['cursor'])
            return friend_list
        elif cmd_args[3] == "a":
            return twitter.users.show(screen_name=cmd_args[4])
        elif cmd_args[3] == "b":
            tweet_data = []
            twitter.blocks.create(screen_name=cmd_args[4])
        elif cmd_args[3] == "ub":
            tweet_data = []
            twitter.blocks.destroy(screen_name=cmd_args[4])
        elif cmd_args[3] == "blocks":
            tweet_data = twitter.blocks.list(skip_status=True)
            block_list = list()
            for user in tweet_data['users']:
                block_list.append(user['screen_name'])
            return block_list
        elif cmd_args[3] == "fav":
            tweet_data = [twitter.favorites.create(_id=cmd_args[4])]
        elif cmd_args[3] == "unfav":
            tweet_data = [twitter.favorites.destroy(_id=cmd_args[4])]
        elif cmd_args[3] == "favs":
            if len(cmd_args) >= 5:
                kwargs = dict()
                if not cmd_args[4].isdigit():
                    kwargs['screen_name'] = cmd_args[4]
                    cmd_args.pop(4)
                if len(cmd_args) == 5:
                    if int(cmd_args[4]) <= 200:
                        kwargs['count'] = int(cmd_args[4])
                    else:
                        kwargs['max_id'] = cmd_args[4]
                elif len(cmd_args) == 6:
                    kwargs['count'] = int(cmd_args[5])
                    kwargs['max_id'] = cmd_args[4]
                tweet_data = twitter.favorites.list(**kwargs)
            else:
                tweet_data = twitter.favorites.list()
        elif cmd_args[3] == "fart":
            twitter.favorites.create(_id=cmd_args[4])
            tweet_data = [twitter.statuses.retweet._(cmd_args[4])()]
        elif cmd_args[3] == "limits":
            output = ""
            if len(cmd_args) >= 5:
                tweet_data = twitter.application.rate_limit_status(
                    resources=", ".join(cmd_args[4:]))
            else:
                tweet_data = twitter.application.rate_limit_status()
            for res in tweet_data['resources']:
                output += res + ":\n"
                for sub_res in tweet_data['resources'][res]:
                    output += "  " + sub_res[len(res)+2:] + ":\n"
                    output += "    " + 'reset' + ": " + \
                    time.strftime('%Y-%m-%d %H:%M:%S',
                                  time.localtime(
                                      tweet_data['resources'][res][sub_res]['reset'])) + "\n"
                    output += "    " + 'limit' + ": " + \
                            str(tweet_data['resources'][res][sub_res]['limit']) + "\n"
                    output += "    " + 'remaining' + ": " + \
                            str(tweet_data['resources'][res][sub_res]['remaining']) + "\n"
            return output
        elif cmd_args[3] == "home":
            if len(cmd_args) == 6:
                kwargs = dict(count=int(cmd_args[5]), max_id=cmd_args[4],
                              exclude_replies=no_home_replies, tweet_mode='extended')
                tweet_data = twitter.statuses.home_timeline(**kwargs)
            elif len(cmd_args) == 5:
                if int(cmd_args[4]) <= 200:
                    tweet_data = twitter.statuses.home_timeline(
                        count=int(cmd_args[4]),
                        exclude_replies=no_home_replies,
                        tweet_mode='extended')
                else:
                    tweet_data = twitter.statuses.home_timeline(max_id=cmd_args[4],
                                                                exclude_replies=no_home_replies,
                                                                tweet_mode='extended')
            else:
                tweet_data = twitter.statuses.home_timeline(exclude_replies=no_home_replies,
                                                            tweet_mode='extended')
        else:
            return "Invalid command: " + cmd_args[3]
    except TwitterHTTPError as err:
        #See if we can print a pretty error message first
        data = err.response_data
        if "errors" in data and "message" in data["errors"][0]:
            return "Error: " + data["errors"][0]["message"]
        return "Unexpected error in get_twitter_data:%s\n Call: %s" % (sys.exc_info(), cmd_args[3])
    except:
        return "Unexpected error in get_twitter_data:%s\n Call: %s" % (sys.exc_info(), cmd_args[3])

    return trim_tweet_data(tweet_data, screen_name, alt_rt_style, is_py3)

# callback for data received in input
def buffer_input_cb(data, buffer, input_data):
    # ...
    end_message = ""
    options = [script_options['screen_name']]
    if script_options['alt_rt_style']:
        options.append("alt_rt_style")
    if script_options['home_replies']:
        options.append("home_replies")
    if is_py3:
        options.append("is_py3")

    if input_data[0] == ':':
        if data != "silent":
            weechat.prnt_date_tags(buffer, 0, "no_highlight", input_data)
        input_args = input_data.split()
        command = input_args[0][1:]
        if command_dict.get(command):
            input_data = input_data.replace(command, command_dict[command], 1)
            command = command_dict[command]
        if command == 'd' and tweet_dict.get(input_args[1]):
            input_data = 'd ' + tweet_dict[input_args[1]]
            weechat.prnt(buffer, "%sYou deleted the following tweet:" % weechat.prefix("network"))
        elif command == 'v' and tweet_dict.get(input_args[1]):
            input_data = 'v ' + tweet_dict[input_args[1]]
            end_message = "Done"
        elif command == 'th' and tweet_dict.get(input_args[1]):
            input_data = 'th ' + tweet_dict[input_args[1]]
            end_message = "Done"
        elif command == 'l' and tweet_dict.get(input_args[1]):
            input_data = 'l ' + tweet_dict[input_args[1]]
        elif command == 'rt' and tweet_dict.get(input_args[1]):
            end_message = "id"
            input_data = 'rt ' + tweet_dict[input_args[1]]
        elif command == 're' and tweet_dict.get(input_args[1]):
            end_message = "id"
            len_id = len(tweet_dict['cur_index'])
            input_data = 're ' + tweet_dict[input_args[1]] \
                    + " '" + html_escape(input_data[(2+1+len_id+1):]) + "'"
        elif command == 'new':
            end_message = "id"
            if script_options['last_id'] != "":
                input_data = 'new ' + script_options['last_id']
            else:
                input_data = 'home'
        elif command == 'home' or command == 'r' or (
                command == 'favs' and len(input_args) >= 2 and input_args[1].isdigit()):
            input_data = command
            if len(input_args) == 3 and tweet_dict.get(input_args[1]) and input_args[2].isdigit():
                num = int(input_args[2])
                # 200 tweets is the max request limit
                if num <= 200 and num > 0:
                    input_data += " " + tweet_dict[input_args[1]] + " " + input_args[2]
                else:
                    input_data += " " + tweet_dict[input_args[1]]
            elif len(input_args) == 2:
                if tweet_dict.get(input_args[1]):
                    input_data += " " + tweet_dict[input_args[1]]
                elif input_args[1].isdigit():
                    num = int(input_args[1])
                    # 200 tweets is the max request limit
                    if num <= 200 and num > 0:
                        input_data += " " + input_args[1]
            end_message = "Done"
        elif command == 'u' or (command == 'favs' and len(input_args) >= 3):
            input_data = " ".join(input_args[:2])[1:]
            if len(input_args) == 4 and tweet_dict.get(input_args[2]) and input_args[3].isdigit():
                num = int(input_args[3])
                # 200 tweets is the max request limit
                if num <= 200 and num > 0:
                    input_data += " " + tweet_dict[input_args[2]] + " " + input_args[3]
                else:
                    input_data += " " + tweet_dict[input_args[2]]
            elif len(input_args) == 3:
                if tweet_dict.get(input_args[2]):
                    input_data += " " + tweet_dict[input_args[2]]
                elif input_args[2].isdigit():
                    num = int(input_args[2])
                    # 200 tweets is the max request limit
                    if num <= 200 and num > 0:
                        input_data += " " + input_args[2]
            end_message = "Done"
        elif command == 'auth':
            if len(input_args) == 2:
                oauth_dance(buffer, input_args[1])
            else:
                oauth_dance(buffer)
            return weechat.WEECHAT_RC_OK
        elif command == 'f' or command == 'fo':
            #L because we are returning a list to be printed later on
            end_message = "L"
            if len(input_args) == 3 and tweet_dict.get(input_args[2]):
                input_data = command + " " + input_args[1] + " " + tweet_dict[input_args[2]]
            elif len(input_args) == 2:
                if tweet_dict.get(input_args[1]):
                    input_data = command + " " + script_options['screen_name'] + " " \
                            + tweet_dict[input_args[1]]
                    #Your list, not any other users
                    end_message += "Y"
                else:
                    input_data = input_data[1:]
            else:
                input_data = command + " " + script_options['screen_name']
                end_message += "Y"
            if command == 'f':
                end_message += "Following"
            else:
                end_message += "Followers"
        elif command == 'a':
            input_data = input_data[1:]
            end_message = "About"
        elif command == 'blocks':
            input_data = input_data[1:]
            end_message = "LBlock list"
        elif command == 'fav' and tweet_dict.get(input_args[1]):
            input_data = 'fav ' + tweet_dict[input_args[1]]
            weechat.prnt(buffer, "%sYou fave'd the following tweet:" % weechat.prefix("network"))
        elif command == 'unfav' and tweet_dict.get(input_args[1]):
            input_data = 'unfav ' + tweet_dict[input_args[1]]
            weechat.prnt(buffer, "%sYou unfave'd the following tweet:" % weechat.prefix("network"))
        elif command == 'fart' and tweet_dict.get(input_args[1]):
            end_message = "id"
            input_data = 'fart ' + tweet_dict[input_args[1]]
            weechat.prnt(buffer, "%sYou fave'd and retweeted the following tweet:" %
                         weechat.prefix("network"))
        elif command == 'cnicks':
            #global tweet_nicks_group
            if tweet_nicks_group[buffer] != "":
                weechat.nicklist_remove_group(buffer, tweet_nicks_group[buffer])
                tweet_nicks_group[buffer] = ""
            tweet_nicks_group[buffer] = weechat.nicklist_add_group(buffer, "",
                                                                   "Tweet_parse",
                                                                   "weechat.color.nicklist_group",
                                                                   1)
            return weechat.WEECHAT_RC_OK
        elif command == 'help':
            weechat.command(buffer, "/help twitter")
            weechat.prnt(buffer, "Exec command /help twitter, check your root buffer")
            return weechat.WEECHAT_RC_OK
        elif command == 'clear':
            weechat.command(buffer, "/buffer clear")
            return weechat.WEECHAT_RC_OK
        elif command == 'stream':
            args = html_escape(input_data[7:])
            weechat.prnt(buffer, create_stream("t_stream", args))
            return weechat.WEECHAT_RC_OK
        elif command == 're_home':
            weechat.prnt(buffer, create_stream("twitter"))
            return weechat.WEECHAT_RC_OK
        else:
            input_data = input_data[1:]
            end_message = "Done"
    else:
        end_message = "id"
        #esacpe special chars when printing to commandline
        input_data = 't ' + "'" + html_escape(input_data) + "'"
        #input_data = 't ' + "'" + html.escape(input_data) + "'"

    weechat.hook_process("python3 " + SCRIPT_FILE_PATH + " " + \
                         script_options["oauth_token"] + " " + script_options["oauth_secret"] + \
                         " " + input_data + "  " + '"' + str(options) + '"', 10 * 1000,
                         "my_process_cb", str([buffer, end_message]))

    return weechat.WEECHAT_RC_OK

def my_command_cb(data, buffer, args):
    # ...

    buffer_input_cb(data, twit_buf, ":"+args)

    return weechat.WEECHAT_RC_OK

def hook_commands_and_completions():
    compl_list = []
    com_list = []
    desc_list = []
    for command in sorted(command_dict):
        compl_list.append(command)
        com_list.append(command + weechat.color("*red") + " or " + \
                        weechat.color('reset') + command_dict[command] + "\n")
        desc_list.append(weechat.color("chat_nick_other") + command + ":    \n" + \
                desc_dict[command])
    weechat.hook_command("twitter", "Command to interact with the twitter api/plugin",
                         " | ".join(com_list),
                         "You can type all of these command in the twitter buffer if you add a "
                         "':' before the command, IE:\n"
                         ":limits\n\n"
                         "If you don't type a command in the twitter buffer you will tweet that "
                         "instead, \n"
                         "text after 280 chars will turn red to let you know were twitter will "
                         "cut off your tweet.\n\n" + weechat.color("*red") + "NOTE:\n"
                         "There are limits on how many twitter api calls you can do, some calls "
                         "are _quite_ restricted.\n"
                         "So if you get HTML errors from the twitter lib you probably exceeded "
                         "the limit\n"
                         "you can check out your limits with the rate_limits/limits command.\n"
                         "_Most_ commands in this plugin only uses one call. If you want to check "
                         "old tweets\n"
                         "in your home timeline it's better to request many tweets in one go.\n"
                         "That way you don't have to request new tweets as often to go further "
                         "back in the timeline.\n"
                         "And thus you are less likely to hit the limit of requests you can do "
                         "in the 15 min time window.\n"
                         "\nYou can write newlines in your tweet with html newline '&#13;&#10;' "
                         "(you can autocomplete it)\n"
                         "\nThe 'number' next to the nicks in the chat window is the <id> of the "
                         "tweet it's used\n"
                         "in the some of the twitter plugin commands.\n\n"
                         "Command desc:\n"+ "\n".join(desc_list),
                         " || ".join(compl_list),
                         "my_command_cb", "")

# callback called when buffer is closed
# TODO rewrite this so it unloads the plugin
def buffer_close_cb(data, buffer):
    # ...
    #TODO handle multiple buffers and free up global buffer pointers
    weechat.unhook_all()
    return weechat.WEECHAT_RC_OK

def tweet_length(message):
    """Replace URLs with placeholders, 20 for http URLs, 21 for https."""
    # regexes to match URLs
    octet = r'(?:2(?:[0-4]\d|5[0-5])|1\d\d|\d{1, 2})'
    ip_addr = r'%s(?:\.%s){3}' % (octet, octet)
    # Base domain regex off RFC 1034 and 1738
    label = r'[0-9a-z][-0-9a-z]*[0-9a-z]?'
    domain = r'%s(?:\.%s)*\.[a-z][-0-9a-z]*[a-z]?' % (label, label)
    url_re = re.compile(r'(\w+://(?:%s|%s)(?::\d+)?(?:/[^\])>\s]*)?)' % (domain, ip_addr), re.I)

    new_message = message

    for url in url_re.findall(message):
        short_url = 'x' * 20
        if url.startswith('https'):
            short_url = 'x' * 21
        new_message = new_message.replace(url, short_url)

    return len(new_message)

def my_modifier_cb(data, modifier, modifier_data, string):
    if not weechat.current_buffer() in friends_nicks_group:
        return string

    #check if this is a commandline argument
    if string == "" or string[0] == "/":
        return string

    length = tweet_length(string)

    # Subtract local command argument from length
    if string[:3] == ":re":
        #:re a2
        length = length - 6

    if length > 280:
        index = 280 - length
        string = string[:index] + weechat.color("*red") + string[index:]

    return string

def parse_oauth_tokens(result):
    for r in result.split('&'):
        k, v = r.split('=')
        if k == 'oauth_token':
            oauth_token = v
        elif k == 'oauth_token_secret':
            oauth_token_secret = v
    return oauth_token, oauth_token_secret

def oauth_proc_cb(data, command, rc, out, err):
    #global script_options
    buffer = twit_buf

    if rc == weechat.WEECHAT_HOOK_PROCESS_ERROR:
        weechat.prnt("", "Error with command '%s'" %
                     command.replace(script_options["oauth_token"], "<token>").replace(
                         script_options["oauth_secret"], "<secret>"))
        return weechat.WEECHAT_RC_OK

    if len(out) > 16 and out[:16] == "Unexpected error":
        #Auth command failed (most likely connection problems)
        weechat.prnt("", out)
        weechat.prnt(buffer, "An error occured, check root buffer. Most likely a connection " +
                     "problem, fix it and the reload weetweet")
        return weechat.WEECHAT_RC_OK

    if out != "":
        if data == "nick":
            weechat.config_set_plugin('screen_name', out.strip())
            finish_init()
        elif data == "friends":
            process_output = ast.literal_eval(out)
            if isinstance(process_output[-1], int):
                t_id = dict_tweet(str(process_output[-1])) + "\t"
                process_output = process_output[:-1]
                weechat.prnt_date_tags(
                    buffer,
                    0,
                    "no_highlight",
                    t_id + "It seems like you are following more than 250 people."
                    "Due to twitter api limits it is nearly impossible to get "
                    "large groups of followers in one go. However the "
                    "nicks will be added when they tweet something so if you don't "
                    "have to be able autocomplete them from the start this is not a "
                    "problem for you. If you want to get the rest of the nicks you can "
                    "use the id of this text.")

            for nick in process_output:
                add_to_nicklist(buffer, nick)
        elif data == "auth1":
            #First auth step to request pin code
            oauth_token, oauth_token_secret = parse_oauth_tokens(out)
            script_options['oauth_token'] = oauth_token
            script_options['oauth_secret'] = oauth_token_secret
            weechat.prnt(buffer, """
    Copy the PIN number that appears on the linked web page and type ":auth <pin>"
    in weechat. For example ":auth 123456"
    """)
            oauth_url = ('https://api.twitter.com/oauth/authorize?oauth_token=' + oauth_token)
            weechat.prnt(buffer, " Please go here to get your PIN: " + oauth_url)
        elif data == "auth2":
            oauth_token, oauth_token_secret = parse_oauth_tokens(out)
            weechat.config_set_plugin('oauth_token', oauth_token)
            weechat.config_set_plugin('oauth_secret', oauth_token_secret)
            weechat.config_set_plugin('auth_complete', "on")
            weechat.prnt(buffer, " Done! now you can begin using this script!")
            weechat.hook_process("python3 " + SCRIPT_FILE_PATH + " " + \
                                 script_options["oauth_token"] + " " + \
                                 script_options["oauth_secret"] + " " + \
                                 "settings []", 10 * 1000, "oauth_proc_cb", "nick")
    return weechat.WEECHAT_RC_OK

def oauth_dance(buffer, pin=""):
    #Auth the twitter client
    if pin == "":
        weechat.prnt(buffer, "Hi there! We're gonna get you all set up to use this plugin.")
        weechat.hook_process("python3 " + SCRIPT_FILE_PATH + " " + "'' " + "'' " + "auth",
                             10 * 1000, "oauth_proc_cb", "auth1")
    else:
        oauth_verifier = pin.strip()
        weechat.hook_process("python3 " + SCRIPT_FILE_PATH + " " + \
                             script_options["oauth_token"] + " " + \
                             script_options["oauth_secret"] + \
                             " " + "auth "+ oauth_verifier, 10 * 1000, "oauth_proc_cb", "auth2")

def setup_buffer(buffer):
    # set title
    weechat.buffer_set(buffer, "title", "Twitter buffer, type ':help' for options.")

    # disable logging, by setting local variable "no_log" to "1"
    weechat.buffer_set(buffer, "localvar_set_no_log", "1")

    #create main nicklist
    friends_nicks_group[buffer] = weechat.nicklist_add_group(buffer, "", "Friends",
                                                             "weechat.color.nicklist_group", 1)

    tweet_nicks_group[buffer] = weechat.nicklist_add_group(buffer, "", "Tweet_parse",
                                                           "weechat.color.nicklist_group", 1)
    autocomp_group = weechat.nicklist_add_group(buffer, "", "Autocomp",
                                                "weechat.color.nicklist_group", 1)
    #newline autocomplete
    weechat.nicklist_add_nick(buffer, autocomp_group, "&#13;&#10;", 'bar_fg', '', '', 1)

    #show nicklist
    weechat.buffer_set(buffer, "nicklist", "1")

    user_nick = script_options['screen_name']

    weechat.buffer_set(buffer, "localvar_set_nick", user_nick)

    add_to_nicklist(buffer, user_nick)
    # Highlight user nick
    weechat.buffer_set(buffer, "highlight_words", user_nick)

def finish_init():
    buffer = twit_buf

    if script_options['screen_name'] == "":
        weechat.hook_process("python3 " + SCRIPT_FILE_PATH + " " + \
                             script_options["oauth_token"] + " " + \
                             script_options["oauth_secret"] + " " + \
                             "settings []", 10 * 1000, "oauth_proc_cb", "nick")
        return
    setup_buffer(buffer)

    #Add friends to nick list
    weechat.hook_process("python3 " + SCRIPT_FILE_PATH + " " + \
                         script_options["oauth_token"] + " " + \
                         script_options["oauth_secret"] + " " + \
                         "f " + script_options['screen_name'] + " []", 10 * 1000,
                         "oauth_proc_cb", "friends")

if __name__ == "__main__" and weechat_call:
    weechat.register(SCRIPT_NAME, "DarkDefender", "1.2.9", "GPL3", "Weechat twitter client", "", "")

    if not import_ok:
        weechat.prnt("", "Can't load twitter python lib >= " + required_twitter_version)
        weechat.prnt("", "Install it via your package manager or go to "
                     "http://mike.verdone.ca/twitter/")
    else:
        hook_commands_and_completions()

        #Set register script options if not available

        for option, value in list(script_options.items()):
            default_value = value[0]
            description = value[1]
            if isinstance(default_value, bool):
                if default_value:
                    default_value = "on"
                else:
                    default_value = "off"
            weechat.config_set_desc_plugin(option,
                                           description + ' (default: "' + default_value + '")')
            if not weechat.config_is_set_plugin(option):
                weechat.config_set_plugin(option, default_value)
                script_options[option] = default_value

        read_config()
        #Hook for config changes
        weechat.hook_config("plugins.var.python." + SCRIPT_NAME + ".*", "config_cb", "")

        #Create buffer
        twit_buf = weechat.buffer_new("twitter", "buffer_input_cb", "", "buffer_close_cb", "")

        #Hook text input so we can update the bar item
        weechat.hook_modifier("input_text_display", "my_modifier_cb", "")

        if script_options['auth_complete']:
            finish_init()
            #Create home_timeline stream.
            #This will indirectly trigger the ":new" command if stream sucessfully starts
            create_stream("twitter")
        else:
            weechat.prnt(twit_buf, """You have to register this plugin with twitter for it to work.
Type ":auth" and follow the instructions to do that""")

elif import_ok:
    if sys.argv[1] == "stream":
        print(twitter_stream(sys.argv))
    else:
        print(get_twitter_data(sys.argv))
else:
    print("Can't load twitter python lib >= " + required_twitter_version)

# vim: tw=0 colorcolumn=100



# -*- coding: utf-8 -*-
#
# Copyright (c) 2012-2017 by nils_2 <weechatter@arcor.de>
# Copyright (c) 2006 by EgS <i@egs.name>
#
# script to keep your nick and recover it in case it's occupied
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# 2017-10-19: nils_2 (freenode.#weechat)
#       1.6 : fix parsing error, now using weechat.info_get_hashtable() (reported by Mikaela)
#
# 2017-10-14: Kaijo & nils_2 (freenode.#weechat)
#       1.5 : fix empty string breaks output
#           : add evaluation for option "text" and use variable "$server" instead of "%s"
#
# 2017-09-06: nils_2 (freenode.#weechat)
#      1.4.2: fix missing weechat.config_string()
#
# 2017-08-17: nils_2 (freenode.#weechat)
#      1.4.1: fix eval_expression for nicks
#           : add evaluation for options
#
# 2017-08-17: nils_2 (freenode.#weechat)
#       1.4 : eval_expression for nicks
#           : use irc.server.<servername>.password
#           : add short /help
#
# 2016-05-12: picasso (freenode.#weechat)
#       1.3 : monitor quits and nick changes
#
# 2014-10-30: nils_2 (freenode.#weechat)
#       1.2 : support of regular expression for server option (idea by michele)
#
# 2013-09-17: nils_2 (freenode.#weechat)
#       1.1 : format of weechat_string_eval_expression() changed
#
# 2013-07-29: nils_2 (freenode.#weechat)
#       1.0 : support of /secure for passwords
#
# 2013-01-29: nils_2 (freenode.#weechat)
#       0.9 : script optimized
#
# 2013-01-27: nils_2 (freenode.#weechat)
#       0.8 : make script compatible with Python 3.x
#
# 2012-10-05: nils_2, (freenode.#weechat)
#       0.7 : add options "plugins.var.python.keepnick.nickserv", "plugins.var.python.keepnick.<server>.password"
#           : changed default "delay" value from 60 to 600 seconds.
#
# 2012-10-04: nils_2, (freenode.#weechat)
#       0.6 : fix bug with case-sensitive nicks (reported by Faethor)
#
# 2012-02-08: nils_2, (freenode.#weechat)
#       0.5 : sync with 0.3.x API (requested by CAHbI4)
#
# requires: WeeChat version 1.3
#
# Development is currently hosted at
# https://github.com/weechatter/weechat-scripts
#
#######################################################################
#                                                                     #
# This script enables the user to keep their nicks and recover it in  #
# case it get's stolen. It uses the servers prefered nicks, so there  #
# is no need for any kind of setup.                                   #
#                                                                     #
# Name:    Keepnick                                                   #
# Licence: GPL v2                                                     #
# Author:  Marcus Eggenberger <i@egs.name>                            #
#
#  Changelog:                                                         #
#   0.4: now starts on load and features user defined check intervals #
#   0.3: Fixed major bug with continuous nickchanges                  #
#   0.2: Fixed Bug: now only checks connected servers                 #
#   0.1: first version released                                       #
#                                                                     #
#######################################################################

try:
    from string import Template
    import weechat,sys,re


except Exception:
    print("This script must be run under WeeChat.")
    print("Get WeeChat now at: http://www.weechat.org/")
    quit()

# -------------------------------[ Constants ]-------------------------------------
SCRIPT_NAME     = "keepnick"
SCRIPT_AUTHOR   = "nils_2 <weechatter@arcor.de>"
SCRIPT_VERSION  = "1.6"
SCRIPT_LICENCE  = "GPL3"
SCRIPT_DESC     = "keep your nick and recover it in case it's occupied"

ISON            = '/ison %s'

OPTIONS         =       { 'delay'       : ('600','delay (in seconds) to look at occupied nick (0 means OFF). It is not recommended to flood the server with /ison requests)'),
                          'timeout'     : ('60','timeout (in seconds) to wait for an answer from server.'),
                          'serverlist'  : ('','comma separated list of servers to look at. Try to register a nickname on server (see: /msg NickServ help).regular expression are allowed (eg. ".*" = matches ALL server,"freen.*" = matches freenode, freenet....) (this option is evaluated).'),
                          'text'        : ('Nickstealer left Network: $server!','text to display, when you get your nick back. (\"$server and $nick\" can be used) (this option is evaluated).'),
                          'nickserv'    : ('/msg -server $server NICKSERV IDENTIFY $passwd','Use SASL authentification, if possible. This command will be used to IDENTIFY you on server (following placeholder can be used: \"$server\" for servername; \"$passwd\" for password). You can create an option for every server to store password: \"plugins.var.python.%s.<servername>.password\", otherwise the \"irc.server.<servername>.password\" option will be used (this option is evaluated).' % SCRIPT_NAME),
                          'command'     : ('/nick %s','This command will be used to rename your nick (\"%s\" will be replaced with your nickname)'),
                          'debug'       : ('off', 'When enabled, will output verbose debugging information during script operation'),
                        }
HOOK            =       { 'timer': '', 'redirect': '', 'quit': '', 'nick': '' }


# ================================[ redirection ]===============================
# calling /ison all x seconds using hook:timer()
def ison(servername,nick,nicklist):
    command = ISON % ' '.join(nicklist)
    debug_print("Checking nicks on server %s with command: %s" % (servername, command) )
    weechat.hook_hsignal_send('irc_redirect_command',
                                  { 'server': servername, 'pattern': 'ison', 'signal': SCRIPT_NAME, 'count': '1', 'string': servername, 'timeout': OPTIONS['timeout'], 'cmd_filter': '' })
    weechat.hook_signal_send('irc_input_send', weechat.WEECHAT_HOOK_SIGNAL_STRING, '%s;;;;%s' % (servername,command))

def redirect_isonhandler(data, signal, hashtable):
    debug_print("ISON response: %s" % hashtable['output'])
    if hashtable['output'] == '':
        return weechat.WEECHAT_RC_OK

    parsed = weechat.info_get_hashtable( "irc_message_parse",dict(message=hashtable['output']) )
    # variable ISON_nicks contains online nicks on server (separated with space)
    # nicks in variable ISON_nicks are lowercase and 'text' contains the nick
    ISON_nicks = [ nick.lower() for nick in parsed['text'].split() ]

    for nick in server_nicks(hashtable['server']):
        mynick = weechat.info_get('irc_nick',hashtable['server'])           # current nick on server

        if nick.lower() == mynick.lower():
            debug_print("I already have nick %s; not changing" % mynick)
            return weechat.WEECHAT_RC_OK
        elif nick.lower() not in ISON_nicks and nick != '':
            grabnick_and_auth(hashtable['server'], nick)
            return weechat.WEECHAT_RC_OK
    return weechat.WEECHAT_RC_OK

# ================================[ functions ]===============================
# nicks used on server
def server_nicks(servername):
    infolist = weechat.infolist_get('irc_server','',servername)
    weechat.infolist_next(infolist)
    nicks = string_eval_expression( weechat.infolist_string(infolist, 'nicks') )    # nicks in config
    weechat.infolist_free(infolist)
    return nicks.split(',')

def server_enabled(servername):
    serverlist = string_eval_expression( OPTIONS['serverlist'] ).split(',')

    server_matched = re.search(r"\b({})\b".format("|".join(serverlist)),servername)

    if servername in serverlist or server_matched:
        return True
    else:
        return False

def check_nicks(data, remaining_calls):
    infolist = weechat.infolist_get('irc_server','','')

    while weechat.infolist_next(infolist):
        servername = weechat.infolist_string(infolist, 'name')
        ptr_buffer = weechat.infolist_pointer(infolist,'buffer')
        nick = weechat.infolist_string(infolist, 'nick')
        ssl_connected = weechat.infolist_integer(infolist,'ssl_connected')
        is_connected = weechat.infolist_integer(infolist,'is_connected')

        if server_enabled(servername):
            if nick and ssl_connected + is_connected:
                ison(servername,nick,server_nicks(servername))
    weechat.infolist_free(infolist)
    return weechat.WEECHAT_RC_OK

def check_quit(data, signal, signal_data):
    servername = signal.split(',')[0]
    hostmask = signal_data.split(':')[1].split(' ')[0]
    nick = hostmask.split('!')[0]
    nicks = server_nicks(servername)
    if server_enabled(servername) and nick in nicks:
        debug_print("Saw %s quit on %s; checking nick" % (nick, servername))
        ison(servername, nick, nicks)
    return weechat.WEECHAT_RC_OK

def check_nick_change(data, signal, signal_data):
    servername = signal.split(',')[0]
    hostmask = signal_data.split(':')[1].split(' ')[0]
    nick = hostmask.split('!')[0]
    nicks = server_nicks(servername)

    if server_enabled(servername) and nick in nicks:
        my_nick = my_nick_on_server(servername)
        if my_nick.lower() != nick.lower():
            debug_print("Saw %s change nick on %s; checking nick" %
                        (nick, servername))
            ison(servername, nick, nicks)
        else:
            debug_print("Saw my own nick change on server (%s); ignoring"
                        % nick)
    return weechat.WEECHAT_RC_OK

def my_nick_on_server(servername):
    nick = weechat.info_get("irc_nick", servername)
    return nick

def grabnick_and_auth(servername, nick):
    global OPTIONS

    password = string_eval_expression( weechat.config_get_plugin('%s.password' % servername) )
    if not password:
        password = string_eval_expression( weechat.config_string(weechat.config_get("irc.server.%s.password" % servername)) )

    grabnick(servername, nick)  # get your nick back

    if password != '' and OPTIONS['nickserv'] != '':
        # command stored in "keepnick.nickserv" option
        t = Template(string_eval_expression( OPTIONS['nickserv']) )
        run_msg = t.safe_substitute(server=servername, passwd=password)
        weechat.command('', run_msg)

def string_eval_expression(string):
    return weechat.string_eval_expression(string,{},{},{})

def grabnick(servername, nick):
    if nick and servername:
        if OPTIONS['text']:
            t = Template( string_eval_expression(OPTIONS['text']) )
            text = t.safe_substitute(server=servername, nick=nick)
            weechat.prnt(weechat.current_buffer(), text)
        weechat.command(weechat.buffer_search('irc','%s.%s' % ('server',servername)), OPTIONS['command'] % nick)

# ================================[ weechat hook ]===============================
def install_hooks():
    global HOOK,OPTIONS

    # Should never happen
    if any([HOOK[k] != '' for k in HOOK]):
        return

    if not OPTIONS['delay'] or not OPTIONS['timeout']:
        return

    debug_print("Installing timer with delay %s seconds" % OPTIONS['delay'])
    HOOK['timer'] = weechat.hook_timer(int(OPTIONS['delay']) * 1000, 0, 0, 'check_nicks', '')
    HOOK['redirect'] = weechat.hook_hsignal('irc_redirection_%s_ison' % SCRIPT_NAME, 'redirect_isonhandler', '' )
    HOOK['quit'] = weechat.hook_signal('*,irc_raw_in_quit', 'check_quit', '')
    HOOK['nick'] = weechat.hook_signal('*,irc_raw_in_nick', 'check_nick_change', '')

    for k in HOOK:
        if HOOK[k] == 0:
            weechat.prnt('', "%s: can't enable %s, hook_timer() failed" %
                         (weechat.prefix('error'), SCRIPT_NAME))

def remove_hooks():
    global HOOK

    for k in HOOK:
        if HOOK[k] != '':
            weechat.unhook(HOOK[k])
            HOOK[k] = ''

def debug_print(msg):
    global OPTIONS
    if OPTIONS['debug'] != 'off':
        weechat.prnt('', "%s DEBUG: %s" % (SCRIPT_NAME, msg))

# ================================[ weechat options and description ]===============================
def init_options():
    global HOOK,OPTIONS
    for option,value in list(OPTIONS.items()):
        weechat.config_set_desc_plugin(option, '%s (default: "%s")' % (value[1], value[0]))
        if not weechat.config_is_set_plugin(option):
            weechat.config_set_plugin(option, value[0])
            OPTIONS[option] = value[0]
        else:
            OPTIONS[option] = weechat.config_get_plugin(option)

def toggle_refresh(pointer, name, value):
    global HOOK,OPTIONS
    option = name[len('plugins.var.python.' + SCRIPT_NAME + '.'):]              # get optionname
    OPTIONS[option] = value                                                     # save new value

    if option == 'delay' or option == 'timeout':
        if int(OPTIONS['delay']) > 0 or int(OPTIONS['timeout']) > 0:
            remove_hooks()
            install_hooks()
        else:
            remove_hooks()                                                  # user switched timer off
    return weechat.WEECHAT_RC_OK

def print_usage(data, buffer, args):
    weechat.prnt(buffer, "%s\t%s: script already running..." % ( string_eval_expression(weechat.config_string(weechat.config_get("weechat.look.prefix_error"))), SCRIPT_NAME) )
    return weechat.WEECHAT_RC_OK

# ================================[ main ]===============================
if __name__ == '__main__':
    if weechat.register(SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION, SCRIPT_LICENCE, SCRIPT_DESC, '',''):
        weechat.hook_command(SCRIPT_NAME,SCRIPT_DESC,
                        '',
                        'You have to edit options with: /set *keepnick*\n'
                        'I suggest using /iset script or /fset plugin.\n',
                        '',
                        'print_usage',
                        '')

        version = weechat.info_get("version_number", "") or 0
        if int(version) >= 0x01030000:
            if int(OPTIONS['delay'][0]) > 0 and int(OPTIONS['timeout'][0]) > 0:
                init_options()
                install_hooks()
                weechat.hook_config( 'plugins.var.python.' + SCRIPT_NAME + '.*', 'toggle_refresh', '' )
        else:
            weechat.prnt('','%s%s %s' % (weechat.prefix('error'),SCRIPT_NAME,': needs version 1.3 or higher'))
            weechat.command('','/wait 1ms /python unload %s' % SCRIPT_NAME)



#
# Copyright (C) 2014 Guido Berhoerster <guido+weechat@berhoerster.name>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import sys
import time
import re
import select
import signal
import errno
import fcntl
import cgi
import multiprocessing


SCRIPT_NAME = 'notification'
APPLICATION = 'Weechat'
VERSION = '1'
AUTHOR = 'Guido Berhoerster'
COPYRIGHT = '(C) 2014 Guido Berhoerster'
SUBTITLE = 'Notification Plugin for Weechat'
HOMEPAGE = 'https://code.guido-berhoerster.org/addons/weechat-scripts/weechat-notification-script/'
EMAIL = 'guido+weechat@berhoerster.name'
DESCRIPTION = 'Notifies of a number of events through desktop notifications ' \
        'and an optional status icon'
DEFAULT_SETTINGS = {
    'status_icon': ('weechat', 'path or name of the status icon'),
    'notification_icon': ('weechat', 'path or name of the icon shown in '
            'notifications'),
    'preferred_toolkit': ('', 'preferred UI toolkit'),
    'notify_on_displayed_only': ('on', 'only notify of messages that are '
            'actually displayed'),
    'notify_on_privmsg': ('on', 'notify when receiving a private message'),
    'notify_on_highlight': ('on', 'notify when a messages is highlighted'),
    'notify_on_dcc_request': ('on', 'notify on DCC requests')
}
BUFFER_SIZE = 1024


class NetstringParser(object):
    """Netstring Stream Parser"""

    IN_LENGTH = 0
    IN_STRING = 1

    def __init__(self, on_string_complete):
        self.on_string_complete = on_string_complete
        self.length = 0
        self.input_buffer = ''
        self.state = self.IN_LENGTH

    def parse(self, data):
        self.input_buffer += data
        ret = True
        while ret:
            if self.state == self.IN_LENGTH:
                ret = self.parse_length()
            else:
                ret = self.parse_string()

    def parse_length(self):
        length, delimiter, self.input_buffer = self.input_buffer.partition(':')
        if not delimiter:
            return False
        try:
            self.length = int(length)
        except ValueError:
            raise SyntaxError('Invalid length: %s' % length)
        self.state = self.IN_STRING
        return True

    def parse_string(self):
        input_buffer_len = len(self.input_buffer)
        if input_buffer_len < self.length + 1:
            return False
        string = self.input_buffer[0:self.length]
        if self.input_buffer[self.length] != ',':
            raise SyntaxError('Missing delimiter')
        self.input_buffer = self.input_buffer[self.length + 1:]
        self.length = 0
        self.state = self.IN_LENGTH
        self.on_string_complete(string)
        return True


def netstring_encode(*args):
    return ''.join(['%d:%s,' % (len(element), element) for element in
            args])

def netstring_decode(netstring):
    result = []
    def append_result(data):
        result.append(data)
    np = NetstringParser(append_result)
    np.parse(netstring)
    return result

def dispatch_weechat_callback(*args):
    return weechat_callbacks[args[0]](*args)

def create_weechat_callback(method):
    global weechat_callbacks

    method_id = str(id(method))
    weechat_callbacks[method_id] = method
    return method_id


class Notifier(object):
    """Simple notifier which discards all notifications, base class for all
       other notifiers
    """

    def __init__(self, icon):
        flags = fcntl.fcntl(sys.stdin, fcntl.F_GETFL)
        fcntl.fcntl(sys.stdin, fcntl.F_SETFL, flags | os.O_NONBLOCK)

        self.parser = NetstringParser(self.on_command_received)

    def on_command_received(self, raw_command):
        command_args = netstring_decode(raw_command)
        if len(command_args) > 1:
            command = command_args[0]
            args = netstring_decode(command_args[1])
        else:
            command = command_args[0]
            args = []
        getattr(self, command)(*args)

    def notify(self, summary, message, icon):
        pass

    def reset(self):
        pass

    def run(self):
        poll = select.poll()
        poll.register(sys.stdin, select.POLLIN | select.POLLPRI)

        while True:
            try:
                events = poll.poll()
            except select.error as e:
                if e.args and e.args[0] == errno.EINTR:
                    continue
                else:
                    raise e
            for fd, event in events:
                if event & (select.POLLIN | select.POLLPRI):
                    buffer_ = os.read(fd, BUFFER_SIZE)
                    if buffer_ != '':
                        self.parser.parse(buffer_)
                if event & (select.POLLERR | select.POLLHUP | select.POLLNVAL):
                    sys.exit(1)


class Gtk2Notifier(Notifier):
    """GTK 2 notifier based on pygtk and pynotify"""

    def __init__(self, icon):
        super(Gtk2Notifier, self).__init__(icon)

        pynotify.init(APPLICATION)

        gobject.io_add_watch(sys.stdin, gobject.IO_IN | gobject.IO_PRI,
                self.on_input)

        if not icon:
            icon_name = None
            icon_pixbuf = None
        elif icon.startswith('/'):
            icon_name = None
            try:
                icon_pixbuf = gtk.gdk.Pixbuf.new_from_file(icon)
            except gobject.GError:
                icon_pixbuf = None
        else:
            icon_name = icon
            icon_pixbuf = None

        if icon_name or icon_pixbuf:
            self.status_icon = gtk.StatusIcon()
            self.status_icon.set_title(APPLICATION)
            self.status_icon.set_tooltip_text(APPLICATION)
            self.status_icon.connect('activate', self.on_activate)
            if icon_name:
                self.status_icon.set_from_icon_name(icon_name)
            elif icon_pixbuf:
                self.status_icon.set_from_pixbuf(icon_pixbuf)
        else:
            self.status_icon = None

    def on_input(self, fd, cond):
        if cond & (gobject.IO_IN | gobject.IO_PRI):
            try:
                buffer_ = os.read(fd.fileno(), BUFFER_SIZE)
                if buffer_ != '':
                    self.parser.parse(buffer_)
            except EOFError:
                gtk.main_quit()
                return False

        if cond & (gobject.IO_ERR | gobject.IO_HUP):
            gtk.main_quit()
            return False

        return True

    def on_activate(self, widget):
        self.reset()

    def on_notification_closed(self, notification):
        if notification.get_closed_reason() == 2:
            self.reset()

    def notify(self, summary, message, icon):
        if self.status_icon:
            self.status_icon.set_tooltip_text('%s: %s' % (APPLICATION,
                    summary))
            self.status_icon.set_blinking(True)

        if icon and icon.startswith('/'):
            icon_name = None
            try:
                icon_pixbuf = gtk.gdk.Pixbuf.new_from_file(icon)
            except gobject.GError:
                icon_pixbuf = None
        else:
            icon_name = icon
            icon_pixbuf = None

        if 'body-markup' in pynotify.get_server_caps():
            body = cgi.escape(message)
        else:
            body = message

        notification = pynotify.Notification(summary, body, icon_name)
        if icon_pixbuf is not None:
            notification.set_image_from_pixbuf(icon_pixbuf)
        notification.connect('closed', self.on_notification_closed)
        notification.show()

    def reset(self):
        if self.status_icon:
            self.status_icon.set_tooltip_text(APPLICATION)
            self.status_icon.set_blinking(False)

    def run(self):
        gtk.main()


class Gtk3Notifier(Notifier):
    """GTK3 notifier based on GObject Introspection Bindings for GTK 3 and
       libnotify
    """

    def __init__(self, icon):
        super(Gtk3Notifier, self).__init__(icon)

        Notify.init(APPLICATION)

        GLib.io_add_watch(sys.stdin, GLib.IO_IN | GLib.IO_PRI, self.on_input)

        if not icon:
            self.icon_name = None
            self.icon_pixbuf = None
        elif icon.startswith('/'):
            self.icon_name = None
            try:
                self.icon_pixbuf = GdkPixbuf.Pixbuf.new_from_file(icon)
            except GLib.GError:
                self.icon_pixbuf = None
        else:
            self.icon_name = icon
            self.icon_pixbuf = None

        if self.icon_name or self.icon_pixbuf:
            # create blank, fully transparent pixbuf in order to simulate
            # blinking
            self.blank_pixbuf = GdkPixbuf.Pixbuf.new(GdkPixbuf.Colorspace.RGB,
                    True, 8, 22, 22)
            self.blank_pixbuf.fill(0x00)

            self.blink_on = True
            self.blink_timeout_id = None

            self.status_icon = Gtk.StatusIcon.new()
            self.status_icon.set_title(APPLICATION)
            self.status_icon.set_tooltip_text(APPLICATION)
            self.status_icon.connect('activate', self.on_activate)
            self.update_icon()
        else:
            self.status_icon = None

    def on_input(self, fd, cond):
        if cond & (GLib.IO_IN | GLib.IO_PRI):
            try:
                self.parser.parse(os.read(fd.fileno(), BUFFER_SIZE))
            except EOFError:
                Gtk.main_quit()
                return False

        if cond & (GLib.IO_ERR | GLib.IO_HUP):
            Gtk.main_quit()
            return False

        return True

    def on_activate(self, widget):
        self.reset()

    def update_icon(self):
        if not self.blink_on:
            self.status_icon.set_from_pixbuf(self.blank_pixbuf)
        elif self.icon_name:
            self.status_icon.set_from_icon_name(self.icon_name)
        elif self.icon_pixbuf:
            self.status_icon.set_from_pixbuf(self.icon_pixbuf)

    def on_blink_timeout(self):
        self.blink_on = not self.blink_on
        self.update_icon()
        return True

    def on_notification_closed(self, notification):
        if notification.get_closed_reason() == 2:
            self.reset()

    def notify(self, summary, message, icon):
        if self.status_icon:
            self.status_icon.set_tooltip_text('%s: %s' % (APPLICATION,
                    summary))
            if self.blink_timeout_id is None:
                self.blink_timeout_id = GLib.timeout_add(500,
                        self.on_blink_timeout)

        if icon and icon.startswith('/'):
            icon_name = None
            try:
                icon_pixbuf = GdkPixbuf.Pixbuf.new_from_file(icon)
            except GLib.GError:
                icon_pixbuf = None
        else:
            icon_name = icon
            icon_pixbuf = None

        if 'body-markup' in Notify.get_server_caps():
            body = cgi.escape(message)
        else:
            body = message

        notification = Notify.Notification.new(summary, body, icon_name)
        if icon_pixbuf is not None:
            notification.set_image_from_pixbuf(icon_pixbuf)
        notification.connect('closed', self.on_notification_closed)
        notification.show()

    def reset(self):
        if self.status_icon:
            self.status_icon.set_tooltip_text(APPLICATION)
            if self.blink_timeout_id is not None:
                GLib.source_remove(self.blink_timeout_id)
                self.blink_timeout_id = None
                self.blink_on = True
                self.update_icon()

    def run(self):
        Gtk.main()


class Qt4Notifier(Notifier):
    """Qt 4 notifier"""

    def __init__(self, icon):
        super(Qt4Notifier, self).__init__(icon)

        signal.signal(signal.SIGINT, self.on_sigint)

        self.qapplication = QtGui.QApplication([])

        self.readable_notifier = QtCore.QSocketNotifier(sys.stdin.fileno(),
                QtCore.QSocketNotifier.Read)
        self.readable_notifier.activated.connect(self.on_input)
        self.readable_notifier.setEnabled(True)

        if not icon:
            self.icon = None
        elif icon.startswith('/'):
            self.icon = QtGui.QIcon(icon)
        else:
            self.icon = QtGui.QIcon.fromTheme(icon)

        if self.icon:
            # create blank, fully transparent pixbuf in order to simulate
            # blinking
            self.blank_icon = QtGui.QIcon()

            self.blink_on = True
            self.blinking_timer = QtCore.QTimer()
            self.blinking_timer.setInterval(500)
            self.blinking_timer.timeout.connect(self.on_blink_timeout)

            self.status_icon = QtGui.QSystemTrayIcon()
            self.status_icon.setToolTip(APPLICATION)
            self.update_icon()
            self.status_icon.setVisible(True)
            self.status_icon.activated.connect(self.on_activated)
            self.status_icon.messageClicked.connect(self.on_message_clicked)
        else:
            self.status_icon = None

    def on_sigint(self, signo, frame):
        self.qapplication.exit(0)

    def on_input(self, fd):
        try:
            self.parser.parse(os.read(fd, BUFFER_SIZE))
        except EOFError:
            self.qapplication.exit(1)

    def on_activated(self, reason):
        self.reset()

    def on_message_clicked(self):
        self.reset()

    def on_blink_timeout(self):
        self.blink_on = not self.blink_on
        self.update_icon()

    def update_icon(self):
        if not self.blink_on:
            self.status_icon.setIcon(self.blank_icon)
        else:
            self.status_icon.setIcon(self.icon)

    def notify(self, summary, message, icon):
        if self.status_icon:
            self.status_icon.setToolTip('%s: %s' % (APPLICATION,
                    cgi.escape(summary)))
            self.blinking_timer.start()
            if self.status_icon.supportsMessages():
                self.status_icon.showMessage(summary, message,
                        QtGui.QSystemTrayIcon.NoIcon)

    def reset(self):
        if self.status_icon:
            self.blinking_timer.stop()
            self.blink_on = True
            self.update_icon()
            self.status_icon.setToolTip(APPLICATION)

    def run(self):
        sys.exit(self.qapplication.exec_())


class KDE4Notifier(Notifier):
    """KDE 4 notifier based on PyKDE4"""

    def __init__(self, icon):
        super(KDE4Notifier, self).__init__(icon)

        signal.signal(signal.SIGINT, self.on_sigint)

        aboutData = kdecore.KAboutData(APPLICATION.lower(), '',
                kdecore.ki18n(APPLICATION), VERSION, kdecore.ki18n(SUBTITLE),
                kdecore.KAboutData.License_GPL_V3, kdecore.ki18n(COPYRIGHT),
                kdecore.ki18n (''), HOMEPAGE, EMAIL)
        kdecore.KCmdLineArgs.init(aboutData)
        self.kapplication = kdeui.KApplication()

        self.readable_notifier = QtCore.QSocketNotifier(sys.stdin.fileno(),
                QtCore.QSocketNotifier.Read)
        self.readable_notifier.activated.connect(self.on_input)
        self.readable_notifier.setEnabled(True)

        if not icon:
            icon_qicon = None
            icon_name = None
        elif icon.startswith('/'):
            icon_qicon = QtGui.QIcon(icon)
            icon_name = None
        else:
            icon_qicon = None
            icon_name = icon

        if icon_name or icon_pixmap:
            self.status_notifier = kdeui.KStatusNotifierItem(self.kapplication)
            self.status_notifier.setCategory(
                    kdeui.KStatusNotifierItem.Communications)
            if icon_name:
                self.status_notifier.setIconByName(icon_name)
                self.status_notifier.setToolTip(icon_name, APPLICATION,
                        SUBTITLE)
            else:
                self.status_notifier.setIconByPixmap(icon_qicon)
                self.status_notifier.setToolTip(icon_qicon, APPLICATION,
                        SUBTITLE)
            self.status_notifier.setStandardActionsEnabled(False)
            self.status_notifier.setStatus(kdeui.KStatusNotifierItem.Active)
            self.status_notifier.setTitle(APPLICATION)
            self.status_notifier.activateRequested.connect(
                    self.on_activate_requested)
        else:
            self.status_notifier = None

    def on_sigint(self, signo, frame):
        self.kapplication.exit(0)

    def on_input(self, fd):
        try:
            self.parser.parse(os.read(fd, BUFFER_SIZE))
        except EOFError:
            self.kapplication.exit(1)

    def on_activate_requested(self, active, pos):
        self.reset()

    def notify(self, summary, message, icon):
        if self.status_notifier:
            self.status_notifier.setToolTipSubTitle(cgi.escape(summary))
            self.status_notifier.setStatus(
                    kdeui.KStatusNotifierItem.NeedsAttention)

        if icon:
            if icon.startswith('/'):
                pixmap = QtGui.QPixmap.load(icon)
            else:
                pixmap = kdeui.KIcon(icon).pixmap(kdeui.KIconLoader.SizeHuge,
                        kdeui.KIconLoader.SizeHuge)
        else:
            pixmap = QtGui.QPixmap()
        kdeui.KNotification.event(kdeui.KNotification.Notification, summary,
                cgi.escape(message), pixmap)

    def reset(self):
        if self.status_notifier:
            self.status_notifier.setStatus(kdeui.KStatusNotifierItem.Active)
            self.status_notifier.setToolTipTitle(APPLICATION)
            self.status_notifier.setToolTipSubTitle(SUBTITLE)

    def run(self):
        sys.exit(self.kapplication.exec_())


class NotificationProxy(object):
    """Proxy object for interfacing with the notifier process"""

    def __init__(self, preferred_toolkit, status_icon):
        self.script_file = os.path.realpath(__file__)
        self._status_icon = status_icon
        self._preferred_toolkit = preferred_toolkit
        self.notifier_process_hook = None
        self.spawn_timer_hook = None
        self.next_spawn_time = 0.0

        self.spawn_notifier_process()

    @property
    def status_icon(self):
        return self._status_icon

    @status_icon.setter
    def status_icon(self, value):
        self._status_icon = value
        self.terminate_notifier_process()
        self.spawn_notifier_process()

    @property
    def preferred_toolkit(self):
        return self._preferred_toolkit

    @preferred_toolkit.setter
    def preferred_toolkit(self, value):
        self._preferred_toolkit = value
        self.terminate_notifier_process()
        self.spawn_notifier_process()

    def on_notifier_process_event(self, data, command, return_code, output,
            error_output):
        if return_code != weechat.WEECHAT_HOOK_PROCESS_RUNNING:
            if return_code == weechat.WEECHAT_HOOK_PROCESS_ERROR:
                error = '%sfailed to run notifier' % weechat.prefix("error")
            else:
                error = '%snotifier exited with exit status %d' % \
                        (weechat.prefix("error"), return_code)
            if output:
                error += '\nstdout:%s' % output
            if error_output:
                error += '\nstderr:%s' % error_output
            weechat.prnt('', error)
            self.notifier_process_hook = None
            self.spawn_notifier_process()
        return weechat.WEECHAT_RC_OK

    def on_spawn_timer(self, data, remaining):
        self.spawn_timer_hook = None
        if not self.notifier_process_hook:
            self.spawn_notifier_process()
        return weechat.WEECHAT_RC_OK

    def spawn_notifier_process(self):
        if self.notifier_process_hook or self.spawn_timer_hook:
            return

        # do not try to respawn a notifier more than once every ten seconds
        now = time.time()
        if long(self.next_spawn_time - now) > 0:
            self.spawn_timer_hook = \
                    weechat.hook_timer(long((self.next_spawn_time - now) *
                    1000), 0, 1, 'dispatch_weechat_callback',
                    create_weechat_callback(self.on_spawn_timer))
            return

        self.next_spawn_time = now + 10
        self.notifier_process_hook = \
                weechat.hook_process_hashtable(sys.executable, {'arg1':
                self.script_file, 'arg2': self.preferred_toolkit, 'arg3':
                self.status_icon, 'stdin': '1'}, 0,
                'dispatch_weechat_callback',
                create_weechat_callback(self.on_notifier_process_event))

    def terminate_notifier_process(self):
        if self.spawn_timer_hook:
            weechat.unhook(self.spawn_timer_hook)
            self.spawn_timer_hook = None
        if self.notifier_process_hook:
            weechat.unhook(self.notifier_process_hook)
            self.notifier_process_hook = None
        self.next_spawn_time = 0.0

    def send(self, command, *args):
        if self.notifier_process_hook:
            if args:
                weechat.hook_set(self.notifier_process_hook, 'stdin',
                        netstring_encode(netstring_encode(command,
                        netstring_encode(*args))))
            else:
                weechat.hook_set(self.notifier_process_hook, 'stdin',
                        netstring_encode(netstring_encode(command)))

    def notify(self, summary, message, icon):
        self.send('notify', summary, message, icon)

    def reset(self):
        self.send('reset')


class NotificationPlugin(object):
    """Weechat plugin"""

    def __init__(self):
        self.DCC_SEND_RE = re.compile(r':(?P<sender>\S+) PRIVMSG \S+ :'
                r'\x01DCC SEND (?P<filename>\S+) \d+ \d+ (?P<size>\d+)')
        self.DCC_CHAT_RE = re.compile(r':(?P<sender>\S+) PRIVMSG \S+ :'
                r'\x01DCC CHAT ')

        weechat.register(SCRIPT_NAME, AUTHOR, VERSION, 'GPL3', DESCRIPTION, '',
                '')

        for option, (value, description) in DEFAULT_SETTINGS.iteritems():
            if not weechat.config_is_set_plugin(option):
                weechat.config_set_plugin(option, value)
            weechat.config_set_desc_plugin(option, '%s (default: "%s")' %
                    (description, value))

        self.notification_proxy = NotificationProxy(
            weechat.config_get_plugin('preferred_toolkit'),
            weechat.config_get_plugin('status_icon'))

        weechat.hook_print('', 'irc_privmsg', '', 1,
                'dispatch_weechat_callback',
                create_weechat_callback(self.on_message))
        weechat.hook_signal('key_pressed', 'dispatch_weechat_callback',
                create_weechat_callback(self.on_key_pressed))
        weechat.hook_signal('irc_dcc', 'dispatch_weechat_callback',
                create_weechat_callback(self.on_dcc))
        weechat.hook_config('plugins.var.python.%s.*' % SCRIPT_NAME,
                'dispatch_weechat_callback',
                create_weechat_callback(self.on_config_changed))

    def on_message(self, data, buffer, date, tags, displayed, highlight,
            prefix, message):
        if weechat.config_get_plugin('notify_on_displayed_only') == 'on' and \
                int(displayed) != 1:
            return weechat.WEECHAT_RC_OK

        formatted_date = time.strftime('%H:%M', time.localtime(float(date)))
        if 'notify_private' in tags.split(',') and \
                weechat.config_get_plugin('notify_on_privmsg') == 'on':
            summary = 'Private message from %s at %s' % (prefix,
                    formatted_date)
            self.notification_proxy.notify(summary, message,
                    weechat.config_get_plugin('notification_icon'))
        elif int(highlight) == 1 and \
                weechat.config_get_plugin('notify_on_highlight') == 'on':
            summary = 'Highlighted message from %s at %s' % (prefix,
                    formatted_date)
            self.notification_proxy.notify(summary, message,
                    weechat.config_get_plugin('notification_icon'))

        return weechat.WEECHAT_RC_OK

    def on_dcc(self, data, signal, signal_data):
        if weechat.config_get_plugin('notify_on_dcc') != 'on':
            return weechat.WEECHAT_RC_OK

        matches = self.DCC_SEND_RE.match(signal_data)
        if matches:
            summary = 'DCC send request from %s' % matches.group('sender')
            message = 'Filname: %s, Size: %d bytes' % \
                    (matches.group('filename'), int(matches.group('size')))
            self.notification_proxy.notify(summary, message,
                    weechat.config_get_plugin('notification_icon'))
            return weechat.WEECHAT_RC_OK

        matches = self.DCC_CHAT_RE.match(signal_data)
        if matches:
            summary = 'DCC chat request from %s' % matches.group('sender')
            message = ''
            self.notification_proxy.notify(summary, message,
                    weechat.config_get_plugin('notification_icon'))
            return weechat.WEECHAT_RC_OK

        return weechat.WEECHAT_RC_OK

    def on_key_pressed(self, data, signal, signal_data):
        self.notification_proxy.reset()
        return weechat.WEECHAT_RC_OK

    def on_config_changed(self, data, option, value):
        if option.endswith('.preferred_toolkit'):
            self.notification_proxy.preferred_toolkit = value
        elif option.endswith('.status_icon'):
            self.notification_proxy.status_icon = value
        return weechat.WEECHAT_RC_OK


def import_modules(modules):
    for module_name, fromlist in modules:
        if fromlist:
            module = __import__(module_name, fromlist=fromlist)
            for identifier in fromlist:
                globals()[identifier] = getattr(module, identifier)
        else:
            globals()[module_name] = __import__(module_name)

def try_import_modules(modules):
    try:
        import_modules(modules)
    except ImportError:
        sys.exit(1)
    sys.exit(0)


if __name__ == '__main__':
    if sys.argv[0] == '__weechat_plugin__':
        # running as Weechat plugin
        import weechat

        weechat_callbacks = {}

        plugin = NotificationPlugin()
    elif len(sys.argv) == 3:
        # running as the notifier process
        preferred_toolkit = sys.argv[1]
        icon = sys.argv[2]

        # required modules for each toolkit
        toolkits_modules = {
            'gtk3': [
                ('gi.repository', [
                    'GLib',
                    'GdkPixbuf',
                    'Gtk',
                    'Notify'
                ])
            ],
            'gtk2': [
                ('pygtk', []),
                ('gobject', []),
                ('gtk', []),
                ('pynotify', [])
            ],
            'qt4': [
                ('PyQt4', [
                    'QtGui',
                    'QtCore'
                ])
            ],
            'kde4': [
                ('PyQt4', [
                    'QtGui',
                    'QtCore'
                ]),
                ('PyKDE4', [
                    'kdecore',
                    'kdeui'
                ])
            ],
            '': []
        }
        available_toolkits = []
        selected_toolkit = ''

        # find available toolkits by spawning a process for each toolkit which
        # tries to import all required modules and returns an exit status of 1
        # in case of an import error
        for toolkit in toolkits_modules:
            process = multiprocessing.Process(target=try_import_modules,
                    args=(toolkits_modules[toolkit],))
            process.start()
            process.join(3)
            if process.is_alive():
                process.terminate()
                process.join()
            if process.exitcode == 0:
                available_toolkits.append(toolkit)

        # select toolkit based on either explicit preference or the
        # availability of modules and the used desktop environment
        if preferred_toolkit:
            if preferred_toolkit in available_toolkits:
                selected_toolkit = preferred_toolkit
        else:
            if 'KDE_FULL_SESSION' in os.environ:
                # preferred order if running KDE4
                toolkits = ['kde4', 'qt4', 'gtk3', 'gtk2']
            else:
                # preferred order for all other desktop environments
                toolkits = ['gtk3', 'gtk2', 'qt4', 'kde4']
            for toolkit in toolkits:
                if toolkit in available_toolkits:
                    selected_toolkit = toolkit
                    break

        # import required toolkit modules
        import_modules(toolkits_modules[selected_toolkit])

        # run selected notifier
        if selected_toolkit == 'gtk3':
            notifier = Gtk3Notifier(icon)
        elif selected_toolkit == 'gtk2':
            notifier = Gtk2Notifier(icon)
        elif selected_toolkit == 'qt4':
            notifier = Qt4Notifier(icon)
        elif selected_toolkit == 'kde4':
            notifier = KDE4Notifier(icon)
        else:
            notifier = Notifier(icon)
        notifier.run()
    else:
        sys.exit(1)



# -*- coding: utf-8 -*-
# MIT License
#
# Copyright (c) 2017-2019 Matthias Adamczyk
# Copyright (c) 2019 Marco Trevisan
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Automatically hide read buffers and unhide them on new activity.

Requires WeeChat version 1.0 or higher.

Configuration:
    plugins.var.python.buffer_autohide.hide_inactive: Hide inactive buffers (default: "off")
    plugins.var.python.buffer_autohide.hide_private: Hide private buffers (default: "off")
    plugins.var.python.buffer_autohide.unhide_low: Unhide a buffer when a low priority message (like JOIN,
        PART, etc.) has been received (default: "off"),
    plugins.var.python.buffer_autohide.exemptions: An enumeration of buffers that should not become hidden (default: "")
    plugins.var.python.buffer_autohide.keep_open: Keep a buffer open for a short amount of time (default: "off")
    plugins.var.python.buffer_autohide.keep_open_timeout: Timeout in milliseconds for how long a selected buffer should be kept around (default: "60 * 1000")

History:
2017-03-19: Matthias Adamczyk <mail@notmatti.me>
    version 0.1: Initial release
2018-06-28: yeled <yeled@github.com>
    version 0.2: Only skip irc.servers
2018-12-07: Matthias Adamczyk <mail@notmatti.me>
    version 0.3: Add a functionality to define exemptions for certain buffers
2018-12-07: Marco Trevisan <mail@3v1n0.net>
    version 0.4: Keep buffers active for a given time before hide them again if they should
2019-01-31: Trygve Aaberge <trygveaa@gmail.com>
    version 0.5: Support buffers from plugins other than IRC as well

https://github.com/notmatti/buffer_autohide
"""
from __future__ import print_function
import ast
import operator as op
import_ok = True
try:
    import weechat
    from weechat import WEECHAT_RC_OK
except ImportError:
    print("Script must be run under weechat. https://weechat.org")
    import_ok = False


SCRIPT_NAME = "buffer_autohide"
SCRIPT_AUTHOR = "Matthias Adamczyk <mail@notmatti.me>"
SCRIPT_VERSION = "0.5"
SCRIPT_LICENSE = "MIT"
SCRIPT_DESC = "Automatically hide read buffers and unhide them on new activity"
SCRIPT_COMMAND = SCRIPT_NAME

DELIMITER = "|@|"
MINIMUM_BUFFER_LIFE = 500 # How many ms are enough to consider a buffer valid
KEEP_ALIVE_TIMEOUT = 60 * 1000 # How long a selected buffer should be kept around

CURRENT_BUFFER = "0x0" # pointer string representation
CURRENT_BUFFER_TIMER_HOOK = None # Timeout hook reference
KEEP_ALIVE_BUFFERS = {} # {pointer_string_rep: timeout_hook}


def config_init():
    """Add configuration options to weechat."""
    global KEEP_ALIVE_TIMEOUT

    config = {
        "hide_inactive": ("off", "Hide inactive buffers"),
        "hide_private": ("off", "Hide private buffers"),
        "unhide_low": ("off",
            "Unhide a buffer when a low priority message (like JOIN, PART, etc.) has been received"),
        "exemptions": ("", "An enumeration of buffers that should not get hidden"),
        "keep_open": ("off", "Keep a buffer open for a short amount of time"),
        "keep_open_timeout": ("60 * 1000", "Timeout in milliseconds for how long a selected buffer should be kept around"),
    }
    for option, default_value in config.items():
        if weechat.config_get_plugin(option) == "":
            weechat.config_set_plugin(option, default_value[0])
        weechat.config_set_desc_plugin(
            option, '{} (default: "{}")'.format(default_value[1], default_value[0]))

    weechat.hook_config("plugins.var.python.buffer_autohide.keep_open_timeout", "timeout_config_changed_cb", "")
    if weechat.config_is_set_plugin("keep_open_timeout"):
        KEEP_ALIVE_TIMEOUT = eval_expr(weechat.config_get_plugin("keep_open_timeout"))


def eval_expr(expr):
    """Evaluate a mathematical expression.

    >>> eval_expr('2 * 6')
    12
    """
    def evaluate(node):
        # supported operators
        operators = {ast.Add: op.add, ast.Sub: op.sub, ast.Mult: op.mul,
                     ast.Div: op.truediv, ast.Pow: op.pow, ast.BitXor: op.xor,
                     ast.USub: op.neg}
        if isinstance(node, ast.Num): # <number>
            return node.n
        elif isinstance(node, ast.BinOp): # <left> <operator> <right>
            return operators[type(node.op)](evaluate(node.left), evaluate(node.right))
        elif isinstance(node, ast.UnaryOp): # <operator> <operand> e.g., -1
            return operators[type(node.op)](evaluate(node.operand))
        else:
            raise TypeError(node)

    return evaluate(ast.parse(expr, mode='eval').body)


def timeout_config_changed_cb(data, option, value):
    """Set the new keep_alive timeout upon change of the corresponding value in plugins.conf."""
    global KEEP_ALIVE_TIMEOUT
    KEEP_ALIVE_TIMEOUT = eval_expr(value)
    return WEECHAT_RC_OK


def hotlist_dict():
    """Return the contents of the hotlist as a dictionary.

    The returned dictionary has the following structure:
    >>> hotlist = {
    ...     "0x0": {                    # string representation of the buffer pointer
    ...         "count_low": 0,
    ...         "count_message": 0,
    ...         "count_private": 0,
    ...         "count_highlight": 0,
    ...     }
    ... }
    """
    hotlist = {}
    infolist = weechat.infolist_get("hotlist", "", "")
    while weechat.infolist_next(infolist):
        buffer_pointer = weechat.infolist_pointer(infolist, "buffer_pointer")
        hotlist[buffer_pointer] = {}
        hotlist[buffer_pointer]["count_low"] = weechat.infolist_integer(
            infolist, "count_00")
        hotlist[buffer_pointer]["count_message"] = weechat.infolist_integer(
            infolist, "count_01")
        hotlist[buffer_pointer]["count_private"] = weechat.infolist_integer(
            infolist, "count_02")
        hotlist[buffer_pointer]["count_highlight"] = weechat.infolist_integer(
            infolist, "count_03")
    weechat.infolist_free(infolist)
    return hotlist


def on_temporary_active_buffer_timeout(buffer, remaining_calls):
    remove_keep_alive(buffer)
    maybe_hide_buffer(buffer)

    return weechat.WEECHAT_RC_OK


def keep_alive_buffer(buffer):
    remove_keep_alive(buffer)

    if buffer_is_hidable(buffer):
        KEEP_ALIVE_BUFFERS[buffer] = weechat.hook_timer(KEEP_ALIVE_TIMEOUT, 0, 1,
            "on_temporary_active_buffer_timeout", buffer)


def remove_keep_alive(buffer):
    global KEEP_ALIVE_BUFFERS

    if buffer in KEEP_ALIVE_BUFFERS.keys():
        weechat.unhook(KEEP_ALIVE_BUFFERS.pop(buffer))


def switch_current_buffer():
    """Save current buffer and ensure that it's visible, then if the
    buffer is elegible to be hidden, we add it to the list of the buffers
    to be hidden after a delay
    """
    global CURRENT_BUFFER
    global CURRENT_BUFFER_TIMER_HOOK

    previous_buffer = CURRENT_BUFFER
    CURRENT_BUFFER = weechat.current_buffer()

    if previous_buffer == CURRENT_BUFFER:
        return

    if weechat.buffer_get_integer(CURRENT_BUFFER, "hidden") == 1:
        weechat.buffer_set(CURRENT_BUFFER, "hidden", "0")

    if weechat.config_get_plugin("keep_open") != "off":
        if CURRENT_BUFFER_TIMER_HOOK is not None:
            weechat.unhook(CURRENT_BUFFER_TIMER_HOOK)
            CURRENT_BUFFER_TIMER_HOOK = None
            maybe_hide_buffer(previous_buffer)
        else:
            keep_alive_buffer(previous_buffer)

        CURRENT_BUFFER_TIMER_HOOK = weechat.hook_timer(MINIMUM_BUFFER_LIFE, 0, 1,
            "on_current_buffer_is_still_active_timeout", "")
    else:
        maybe_hide_buffer(previous_buffer)


def on_current_buffer_is_still_active_timeout(pointer, remaining_calls):
    global CURRENT_BUFFER_TIMER_HOOK
    global KEEP_ALIVE_BUFFERS

    CURRENT_BUFFER_TIMER_HOOK = None
    remove_keep_alive(CURRENT_BUFFER)

    return weechat.WEECHAT_RC_OK


def switch_buffer_cb(data, signal, signal_data):
    """
    :param data: Pointer
    :param signal: Signal sent by Weechat
    :param signal_data: Data sent with signal
    :returns: callback return value expected by Weechat.
    """
    switch_current_buffer()
    return WEECHAT_RC_OK


def buffer_is_hidable(buffer):
    """Check if passed buffer can be hidden.

    If configuration option ``hide_private`` is enabled,
    private buffers will become hidden as well.

    If the previous buffer name matches any of the exemptions defined in ``exemptions``,
    it will not become hidden.

    :param buffer: Buffer string representation
    """
    if buffer == weechat.current_buffer():
        return False

    if buffer in KEEP_ALIVE_BUFFERS.keys():
        return False

    full_name = weechat.buffer_get_string(buffer, "full_name")

    if full_name.startswith("irc.server"):
        return False

    buffer_type = weechat.buffer_get_string(buffer, 'localvar_type')

    if (buffer_type == "private"
            and weechat.config_get_plugin("hide_private") == "off"):
        return False

    if weechat.config_get_plugin("hide_inactive") == "off":
        nicks_count = weechat.buffer_get_integer(buffer, 'nicklist_nicks_count')
        if nicks_count == 0:
            return False

    for entry in list_exemptions():
        if entry in full_name:
            return False

    return True


def maybe_hide_buffer(buffer):
    """Hide a buffer if all the conditions are met"""
    if buffer_is_hidable(buffer):
        weechat.buffer_set(buffer, "hidden", "1")


def unhide_buffer_cb(data, signal, signal_data):
    """Unhide a buffer on new activity.

    This callback unhides a buffer in which a new message has been received.
    If configuration option ``unhide_low`` is enabled,
    buffers with only low priority messages (like JOIN, PART, etc.) will be unhidden as well.

    :param data: Pointer
    :param signal: Signal sent by Weechat
    :param signal_data: Data sent with signal
    :returns: Callback return value expected by Weechat.
    """
    hotlist = hotlist_dict()
    line_data = weechat.hdata_pointer(weechat.hdata_get('line'), signal_data, 'data')
    buffer = weechat.hdata_pointer(weechat.hdata_get('line_data'), line_data, 'buffer')

    if not buffer in hotlist.keys():
        # just some background noise
        return WEECHAT_RC_OK

    if (weechat.config_get_plugin("unhide_low") == "on"
            and hotlist[buffer]["count_low"] > 0
            or hotlist[buffer]["count_message"] > 0
            or hotlist[buffer]["count_private"] > 0
            or hotlist[buffer]["count_highlight"] > 0):
        remove_keep_alive(buffer)
        weechat.buffer_set(buffer, "hidden", "0")

    return WEECHAT_RC_OK


def list_exemptions():
    """Return a list of exemption defined in ``exemptions``.

    :returns: A list of defined exemptions.
    """
    return [x for x in weechat.config_get_plugin("exemptions").split(DELIMITER) if x != ""]


def add_to_exemptions(entry):
    """Add an entry to the list of exemptions.

    An entry can be either a #channel or server_name.#channel

    :param entry: The entry to add.
    :returns: the new list of entries. The return value is only used for unit testing.
    """
    entries = list_exemptions()
    entries.append(entry)
    weechat.config_set_plugin("exemptions", DELIMITER.join(entries))
    weechat.prnt("", "[{}] add: {} added to exemptions.".format(SCRIPT_COMMAND, entry))
    return entries


def del_from_exemptions(entry):
    """Remove an entry from the list of defined exemptions.

    :param entry: The entry to delete, which can be specified by the position in the list or by the name itself.
    :returns: the new list of entries. The return value is only used for unit testing.
    """
    entries = list_exemptions()
    try:
        # by index
        try:
            index = int(entry) - 1
            if index < 0:
                raise IndexError
            entry = entries.pop(index)
        except IndexError:
            weechat.prnt("", "[{}] del: Index out of range".format(SCRIPT_COMMAND))
            return entries
    except ValueError:
        try:
            # by name
            entries.remove(entry)
            weechat.config_set_plugin("exemptions", DELIMITER.join(entries))
        except ValueError:
            weechat.prnt("", "[{}] del: Could not find {}".format(SCRIPT_COMMAND, entry))
            return entries

    weechat.config_set_plugin("exemptions", DELIMITER.join(entries))
    weechat.prnt("", "[{}] del: Removed {} from exemptions.".format(SCRIPT_COMMAND, entry))
    return entries


def print_exemptions():
    """Print all exemptions defined in ``exemptions``"""
    entries = list_exemptions()
    if entries:
        count = 1
        for entry in entries:
            weechat.prnt("", "[{}] {}: {}".format(SCRIPT_COMMAND, count, entry))
            count += 1
    else:
        weechat.prnt("", "[{}] list: No exemptions defined so far.".format(SCRIPT_COMMAND))


def command_cb(data, buffer, args):
    """Weechat callback for parsing and executing the given command.

    :returns: Callback return value expected by Weechat.
    """
    list_args = args.split(" ")

    if list_args[0] not in ["add", "del", "list"]:
        weechat.prnt("", "[{0}] bad option while using /{0} command, try '/help {0}' for more info".format(
            SCRIPT_COMMAND))

    elif list_args[0] == "add":
        if len(list_args) == 2:
            add_to_exemptions(list_args[1])

    elif list_args[0] == "del":
        if len(list_args) == 2:
            del_from_exemptions(list_args[1])

    elif list_args[0] == "list":
        print_exemptions()
    else:
        weechat.command("", "/help " + SCRIPT_COMMAND)

    return WEECHAT_RC_OK


if (__name__ == '__main__' and import_ok and weechat.register(
            SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION, SCRIPT_LICENSE, SCRIPT_DESC, '', '')):
    weechat_version = weechat.info_get("version_number", "") or 0
    if int(weechat_version) >= 0x01000000:
        config_init()
        CURRENT_BUFFER = weechat.current_buffer()
        weechat.hook_signal("buffer_switch", "switch_buffer_cb", "")
        weechat.hook_signal("buffer_line_added", "unhide_buffer_cb", "")
        weechat.hook_command(
            SCRIPT_NAME,
            SCRIPT_DESC,
            "add $buffer_name | del { $buffer_name | $list_position } | list",
            "  add    : Add $buffer_name to the list of exemptions\n"
            "           $buffer_name can be either #channel or server_name.#channel\n"
            "  del    : Delete $buffer_name from the list of exemptions\n"
            "  list   : Return a list of all buffers that should not become hidden.",
            "add|del|list",
            "command_cb",
            ""
        )
    else:
        weechat.prnt("", "{}{} requires WeeChat version 1.0 or higher".format(
            weechat.prefix('error'), SCRIPT_NAME))



# -*- coding: utf-8 -*-
# Author: ochameau <poirot.alex AT gmail DOT com>
# Homepage: https://github.com/ochameau/weechat-notifo
# Derived from: notify
#   Author: lavaramano <lavaramano AT gmail DOT com>
#   Improved by: BaSh - <bash.lnx AT gmail DOT com>
#   Ported to Weechat 0.3.0 by: Sharn - <sharntehnub AT gmail DOT com)
# And from: notifo_notify
#   Author: SAEKI Yoshiyasu <laclef_yoshiyasu@yahoo.co.jp>
#   Homepage: http://bitbucket.org/laclefyoshi/weechat/
# This plugin send push notifications to your iPhone or Android smartphone
# by using Notifo.com mobile application/services
# Requires Weechat 0.3.0
# Released under GNU GPL v2
#
# 2014-05-10, Sébastien Helleu <flashcode@flashtux.org>
#     version 0.2: change hook_print callback argument type of
#                  displayed/highlight (WeeChat >= 1.0)
#
# 2011-08-27, ochameau <poirot.alex@gmail.com>:
#     version 0.1: merge notify.py and notifo_notify.py in order to avoid
#                  sending notifications when channel or private buffer is
#                  already opened

import weechat, string, urllib, urllib2

weechat.register("notifo", "ochameau", "0.2", "GPL", "notifo: Send push notifications to your iPhone/Android about your private messages and highlights.", "", "")

credentials = {
    "username": "",
    "api_secret": ""
}

for option, default_value in credentials.items():
    if weechat.config_get_plugin(option) == "":
        weechat.prnt("", weechat.prefix("error") + "notifo: Please set option: %s" % option)
        weechat.prnt("", "notifo: /set plugins.var.python.notifo.%s STRING" % option)

# Hook privmsg/hilights
weechat.hook_print("", "irc_privmsg", "", 1, "notify_show", "")

# Functions
def notify_show(data, bufferp, uber_empty, tagsn, isdisplayed,
        ishilight, prefix, message):

    if (bufferp == weechat.current_buffer()):
        pass

    elif weechat.buffer_get_string(bufferp, "localvar_type") == "private":
        show_notification(prefix, message)

    elif int(ishilight):
        buffer = (weechat.buffer_get_string(bufferp, "short_name") or
                weechat.buffer_get_string(bufferp, "name"))
        show_notification(buffer, prefix + ": " + message)

    return weechat.WEECHAT_RC_OK

def show_notification(chan, message):
    NOTIFO_USER = weechat.config_get_plugin("username")
    NOTIFO_API_SECRET = weechat.config_get_plugin("api_secret")
    if NOTIFO_USER != "" and NOTIFO_API_SECRET != "":
        url = "https://api.notifo.com/v1/send_notification"
        opt_dict = {
            "msg": message,
            "label": "weechat",
            "title": chan
            }
        opt = urllib.urlencode(opt_dict)
        basic = "Basic %s" % ":".join([NOTIFO_USER, NOTIFO_API_SECRET]).encode("base64").strip()
        python2_bin = weechat.info_get("python2_bin", "") or "python"
        weechat.hook_process(
            python2_bin + " -c \"import urllib2\n"
            "req = urllib2.Request('" + url + "', '" + opt + "')\n"
            "req.add_header('Authorization', '" + basic + "')\n"
            "res = urllib2.urlopen(req)\n\"",
            30 * 1000, "", "")

# vim: autoindent expandtab smarttab shiftwidth=4



# -*- coding: utf-8 -*-

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# This script checks stream status of any channel on any servers
# listed in the "plugins.var.python.twitch.servers" setting. When you
# switch to a buffer it will display updated infomation about the stream
# in the title bar. Typing '/twitch' in buffer will also fetch updated
# infomation. '/whois nick' will lookup user info and display it in current
# buffer.
#
# https://github.com/mumixam/weechat-twitch
#
# settings:
# plugins.var.python.twitch.servers (default: twitch)
# plugins.var.python.twitch.prefix_nicks (default: 1)
# plugins.var.python.twitch.debug (default: 0)
# plugins.var.python.twitch.ssl_verify (default: 1)
#
# # History:
#
#
# 2019-09-25, mumixam
#     v0.7: updated script to use current api
#
# 2019-03-03,
#     v0.6: added support for CLEARMSG -MentalFS
#           fixed issue with /whois -mumixam
# 2018-06-03, mumixam
#     v0.5: enable curl verbose mode when debug is active, add option to disable ssl/tls verification,
#           if stream title contains newline char replace it with space
# 2017-11-02, mumixam
#     v0.4: added debug mode for API calls, minor bugfixes
# 2017-06-10, mumixam
#     v0.3: fixed whois output of utf8 display names
# 2016-11-03, mumixam
#     v0.2: added detailed /help
# 2016-10-30, mumixam
#     v0.1: script added to weechat.org



SCRIPT_NAME = "twitch"
SCRIPT_AUTHOR = "mumixam"
SCRIPT_VERSION = "0.7"
SCRIPT_LICENSE = "GPL3"
SCRIPT_DESC = "twitch.tv Chat Integration"
OPTIONS={
    'servers': ('twitch','Name of server(s) which script will be active on, space seperated'),
    'prefix_nicks': ('1','Prefix nicks based on ircv3 tags for mods/subs, This can be cpu intensive on very active chats [1 for enabled, 0 for disabled]'),
    'debug': ('0','Debug mode'),
    'ssl_verify': ('1', 'Verify SSL/TLS certs')
}


import weechat
import json
from calendar import timegm
from datetime import datetime, timedelta
import time
import string
import ast

curlopt = {
    "httpheader": "Client-ID: awtv6n371jb7uayyc4jaljochyjbfxs",
    "timeout": "5",
    "verbose": "0",
    "ssl_verifypeer": "1",
    "ssl_verifyhost": "2"
}


gameid_cache = {}
uid_cache = {}

def days_hours_minutes(td):
    age = ''
    hours = td.seconds // 3600
    min = td.seconds // 60 % 60
    if not td.days == 0:
        age += str(td.days) + 'd '
    if not hours == 0:
        age += str(hours) + 'h '
    if not min == 0:
        age += str(min) + 'm'
    return age.strip()


def twitch_main(data, buffer, args):
    if not args == 'bs':
        weechat.buffer_set(buffer, 'localvar_set_tstatus', '')
    username = weechat.buffer_get_string(buffer, 'short_name').replace('#', '')
    server = weechat.buffer_get_string(buffer, 'localvar_server')
    type = weechat.buffer_get_string(buffer, 'localvar_type')
    if not (server in OPTIONS['servers'].split() and type == 'channel'):
        return weechat.WEECHAT_RC_OK
    url = 'https://api.twitch.tv/helix/streams?user_login=' + username
    weechat.hook_process_hashtable(
        "url:" + url, curlopt, 7 * 1000, "stream_api", buffer)
    return weechat.WEECHAT_RC_OK


def makeutf8(data):
    data = data.encode('utf8')
    if not isinstance(data, str):
        data=str(data,'utf8')
    return data


def stream_api(data, command, rc, stdout, stderr):
    try:
        jsonDict = json.loads(stdout.strip())
    except Exception as e:
        weechat.prnt(data, '%stwitch.py: error communicating with twitch api' % weechat.prefix('error'))
        if OPTIONS['debug']:
            weechat.prnt(data,'%stwitch.py: return code: %s' % (weechat.prefix('error'),rc))
            weechat.prnt(data,'%stwitch.py: stdout: %s' % (weechat.prefix('error'),stdout))
            weechat.prnt(data,'%stwitch.py: stderr: %s' % (weechat.prefix('error'),stderr))
            weechat.prnt(data,'%stwitch.py: exception: %s' % (weechat.prefix('error'),e))
        return weechat.WEECHAT_RC_OK
    currentbuf = weechat.current_buffer()
    title_fg = weechat.color(
        weechat.config_color(weechat.config_get("weechat.bar.title.color_fg")))
    title_bg = weechat.color(
        weechat.config_color(weechat.config_get("weechat.bar.title.color_bg")))
    pcolor = weechat.color('chat_prefix_network')
    ccolor = weechat.color('chat')
    red = weechat.color('red')
    blue = weechat.color('blue')
    green = weechat.color('green')
    ptime = time.strftime("%H:%M:%S")
    subs = weechat.buffer_get_string(data, 'localvar_subs')
    r9k = weechat.buffer_get_string(data, 'localvar_r9k')
    slow = weechat.buffer_get_string(data, 'localvar_slow')
    emote = weechat.buffer_get_string(data, 'localvar_emote')
    if not 'data' in jsonDict.keys():
        weechat.prnt(data, 'twitch.py: Error with twitch API (data key missing from json)')
        return weechat.WEECHAT_RC_OK
    if not jsonDict['data']:
        line = "STREAM: %sOFFLINE%s %sCHECKED AT: (%s)" % (
            red, title_fg, blue, ptime)
        if subs:
            line += " %s[SUBS]" % title_fg
        if r9k:
            line += " %s[R9K]" % title_fg
        if slow:
            line += " %s[SLOW@%s]" % (title_fg, slow)
        if emote:
            line += " %s[EMOTE]" % title_fg
        weechat.buffer_set(data, "title", line)
    else:
        currenttime = time.time()
        if len(jsonDict['data']) == 1:
            jsonDict['data'] = jsonDict['data'][0]
        output = 'STREAM: %sLIVE%s' % (green, title_fg)
        if 'game_id' in jsonDict['data']:
            if jsonDict['data']['game_id']:
                game = jsonDict['data']['game_id']
                game_id = game
                if game in gameid_cache:
                    game = gameid_cache[game]
                output += ' <%s> with' % game
        if 'viewer_count' in jsonDict['data']:
            viewers = jsonDict['data']['viewer_count']
            output += ' %s viewers started' % viewers
        if 'started_at' in jsonDict['data']:
            createtime = jsonDict['data']['started_at'].replace('Z', 'GMT')
            starttime = timegm(
                time.strptime(createtime, '%Y-%m-%dT%H:%M:%S%Z'))
            dur = timedelta(seconds=currenttime - starttime)
            uptime = days_hours_minutes(dur)
            output += ' %s ago' % uptime
        if 'title' in jsonDict['data']:
            titleutf8=jsonDict['data']['title'].replace('\n',' ').encode('utf8')
            titleascii=jsonDict['data']['title'].encode('ascii','replace')
            if not isinstance(titleutf8, str):
                titleascii=str(titleascii,'utf8')
                titleutf8=str(titleutf8,'utf8')
            oldtitle = weechat.buffer_get_string(data, 'localvar_tstatus')
            if not oldtitle == titleascii:
                weechat.prnt(data, '%s--%s Title is "%s"' %
                             (pcolor, ccolor, titleutf8))
                weechat.buffer_set(data, 'localvar_set_tstatus', titleascii)

        output += ' (%s)' % ptime
        if subs:
            output += " %s[SUBS]" % title_fg
        if r9k:
            output += " %s[R9K]" % title_fg
        if slow:
            output += " %s[SLOW@%s]" % (title_fg, slow)
        if emote:
            output += " %s[EMOTE]" % title_fg
        weechat.buffer_set(data, "title", output)
        if not game_id in gameid_cache:
            url = 'https://api.twitch.tv/helix/games?id=' + game_id
            weechat.hook_process_hashtable(
                "url:" + url, curlopt, 7 * 1000, "game_api", data)

    return weechat.WEECHAT_RC_OK



def game_api(data, command, rc, stdout, stderr):
    try:
        jsonDict = json.loads(stdout.strip())
    except Exception as e:
        weechat.prnt(data, '%stwitch.py: error communicating with twitch api' % weechat.prefix('error'))
        if OPTIONS['debug']:
            weechat.prnt(data,'%stwitch.py: return code: %s' % (weechat.prefix('error'),rc))
            weechat.prnt(data,'%stwitch.py: stdout: %s' % (weechat.prefix('error'),stdout))
            weechat.prnt(data,'%stwitch.py: stderr: %s' % (weechat.prefix('error'),stderr))
            weechat.prnt(data,'%stwitch.py: exception: %s' % (weechat.prefix('error'),e))
        return weechat.WEECHAT_RC_OK

    if 'data' in jsonDict.keys():
        if not jsonDict['data']:
            return weechat.WEECHAT_RC_OK
        if len(jsonDict['data']) == 1:
            jsonDict['data'] = jsonDict['data'][0]
        old_title = weechat.buffer_get_string(data, "title")
        id = jsonDict['data']['id']
        name = makeutf8(jsonDict['data']['name'])
        new_title = old_title.replace('<{}>'.format(id),'<{}>'.format(name))
        weechat.buffer_set(data, "title", new_title)
        gameid_cache[id] = name
    return weechat.WEECHAT_RC_OK



def channel_api(data, command, rc, stdout, stderr):
    try:
        jsonDict = json.loads(stdout.strip())
    except Exception as e:
        weechat.prnt(data, '%stwitch.py: error communicating with twitch api' % weechat.prefix('error'))
        if OPTIONS['debug']:
            weechat.prnt(data['buffer'],'%stwitch.py: return code: %s' % (weechat.prefix('error'),rc))
            weechat.prnt(data['buffer'],'%stwitch.py: stdout: %s' % (weechat.prefix('error'),stdout))
            weechat.prnt(data['buffer'],'%stwitch.py: stderr: %s' % (weechat.prefix('error'),stderr))
            weechat.prnt(data['buffer'],'%stwitch.py: exception: %s' % (weechat.prefix('error'),e))
        return weechat.WEECHAT_RC_OK
    currentbuf = weechat.current_buffer()
    pcolor = weechat.color('chat_prefix_network')
    ccolor = weechat.color('chat')
    dcolor = weechat.color('chat_delimiters')
    ncolor = weechat.color('chat_nick')
    ul = weechat.color("underline")
    rul = weechat.color("-underline")
    pformat = weechat.config_string(
        weechat.config_get("weechat.look.prefix_network"))
    
    if 'total' in jsonDict:
        uid = command.split('=')[-1]
        name = 'WHOIS'
        if 'to_id' in command:
            followers = jsonDict['total']
            if uid in uid_cache:
                name = uid_cache[uid]
            output = '%s%s %s[%s%s%s]%s %sFollowers%s: %s' % (
                pcolor, pformat, dcolor, ncolor, name, dcolor, ccolor, ul, rul, followers)
            weechat.prnt(data, makeutf8(output))
            url = 'https://api.twitch.tv/helix/users/follows?from_id=' + uid
            url_hook = weechat.hook_process_hashtable(
                "url:" + url, curlopt, 7 * 1000, "channel_api", data)
            return weechat.WEECHAT_RC_OK
        if 'from_id' in command:
            following = jsonDict['total']
            if uid in uid_cache:
                name = uid_cache[uid]
            output = '%s%s %s[%s%s%s]%s %sFollowing%s: %s' % (
                pcolor, pformat, dcolor, ncolor, name, dcolor, ccolor, ul, rul, following)
            weechat.prnt(data, makeutf8(output))
            return weechat.WEECHAT_RC_OK
    if ('users' in jsonDict) and jsonDict['users'] and len(jsonDict['users'][0]) == 8:
        dname = jsonDict['users'][0]['display_name']
        name = jsonDict['users'][0]['name']
        create = jsonDict['users'][0]['created_at'].split('T')[0]
        status = jsonDict['users'][0]['bio']
        uid = jsonDict['users'][0]['_id']
        uid_cache[uid] = name
        output = '%s%s %s[%s%s%s]%s %sDisplay Name%s: %s' % (
                            pcolor, pformat, dcolor, ncolor, name, dcolor, ccolor, ul, rul, dname)
        output += '\n%s%s %s[%s%s%s]%s %sAccount Created%s: %s' % (
            pcolor, pformat, dcolor, ncolor, name, dcolor, ccolor, ul, rul, create)
        if status:
            output += '\n%s%s %s[%s%s%s]%s %sBio%s: %s' % (
                pcolor, pformat, dcolor, ncolor, name, dcolor, ccolor, ul, rul, status)
        weechat.prnt(data, makeutf8(output))
        url = 'https://api.twitch.tv/helix/users/follows?to_id=' + uid
        url_hook = weechat.hook_process_hashtable(
            "url:" + url, curlopt, 7 * 1000, "channel_api", data)

    else:
        weechat.prnt(data, 'Error: No Such User')

    return weechat.WEECHAT_RC_OK


def twitch_clearchat(data, modifier, modifier_data, string):
    mp = weechat.info_get_hashtable(
    'irc_message_parse', {"message": string})
    server = modifier_data
    user = mp['text']
    channel = mp['channel']
    try:
        tags = dict([s.split('=') for s in mp['tags'].split(';')])
    except:
        tags = ''
    buffer = weechat.buffer_search("irc", "%s.%s" % (server, channel))
    if buffer:
        pcolor = weechat.color('chat_prefix_network')
        ccolor = weechat.color('chat')
        ul = weechat.color("underline")
        rul = weechat.color("-underline")
        if user:
            if 'ban-duration' in tags:
                if 'ban-reason' in tags and tags['ban-reason']:
                    bn=tags['ban-reason'].replace('\s',' ')
                    weechat.prnt(buffer,"%s--%s %s has been timed out for %s seconds %sReason%s: %s" %
                        (pcolor, ccolor, user, tags['ban-duration'], ul, rul, bn))
                else:
                    weechat.prnt(buffer,"%s--%s %s has been timed out for %s seconds" %
                        (pcolor, ccolor, user, tags['ban-duration']))
            elif 'ban-reason' in tags:
                if tags['ban-reason']:
                    bn=tags['ban-reason'].replace('\s',' ')
                    weechat.prnt(buffer,"%s--%s %s has been banned %sReason%s: %s" %
                        (pcolor, ccolor, user, ul, rul,bn))
                else:
                    weechat.prnt(buffer,"%s--%s %s has been banned" %
                        (pcolor, ccolor, user))
            else:
                weechat.prnt(
                    buffer, "%s--%s %s's Chat Cleared By Moderator" % (pcolor, ccolor, user))
        else:
            weechat.prnt(
                buffer, "%s--%s Entire Chat Cleared By Moderator" % (pcolor, ccolor))
    return ""


def twitch_clearmsg(data, modifier, modifier_data, string):
    mp = weechat.info_get_hashtable(
    'irc_message_parse', {"message": string})
    server = modifier_data
    channel = mp['channel']
    try:
        tags = dict([s.split('=') for s in mp['tags'].split(';')])
    except:
        tags = ''
    buffer = weechat.buffer_search("irc", "%s.%s" % (server, channel))
    if buffer:
        pcolor = weechat.color('chat_prefix_network')
        ccolor = weechat.color('chat')
        if 'login' in tags:
            weechat.prnt(buffer,"%s--%s a message from %s was deleted" % (pcolor, ccolor, tags['login']))
        else:
            weechat.prnt(buffer, "%s--%s a message was deleted" % (pcolor, ccolor))
    return ""


def twitch_suppress(data, modifier, modifier_data, string):
    return ""


def twitch_reconnect(data, modifier, modifier_data, string):
    server = modifier_data
    buffer = weechat.buffer_search("irc", "server.%s" % server)
    if buffer:
        pcolor = weechat.color('chat_prefix_network')
        ccolor = weechat.color('chat')
        weechat.prnt(
            buffer, "%s--%s Server sent reconnect request. Issuing /reconnect" % (pcolor, ccolor))
        weechat.command(buffer, "/reconnect")
    return ""


def twitch_buffer_switch(data, signal, signal_data):
    server = weechat.buffer_get_string(signal_data, 'localvar_server')
    type = weechat.buffer_get_string(signal_data, 'localvar_type')
    if not (server in OPTIONS['servers'].split() and type == 'channel'):
        return weechat.WEECHAT_RC_OK
    twitch_main('', signal_data, 'bs')
    return weechat.WEECHAT_RC_OK


def twitch_roomstate(data, modifier, server, string):
    message = weechat.info_get_hashtable(
        'irc_message_parse', {"message": string})
    buffer = weechat.buffer_search(
        "irc", "%s.%s" % (server, message['channel']))
    for tag in message['tags'].split(';'):
        if tag == 'subs-only=0':
            weechat.buffer_set(buffer, 'localvar_set_subs', '')
        if tag == 'subs-only=1':
            weechat.buffer_set(buffer, 'localvar_set_subs', '1')
        if tag == 'r9k=0':
            weechat.buffer_set(buffer, 'localvar_set_r9k', '')
        if tag == 'r9k=1':
            weechat.buffer_set(buffer, 'localvar_set_r9k', '1')
        if tag == 'emote-only=0':
            weechat.buffer_set(buffer, 'localvar_set_emote', '')
        if tag == 'emote-only=1':
            weechat.buffer_set(buffer, 'localvar_set_emote', '1')
        if tag.startswith('slow='):
            value = tag.split('=')[-1]
            if value == '0':
                weechat.buffer_set(buffer, 'localvar_set_slow', '')
            if value > '0':
                weechat.buffer_set(buffer, 'localvar_set_slow', value)
        twitch_main('', buffer, 'bs')
    return ''


def twitch_usernotice(data, modifier, server, string):
    pcolor = weechat.color('chat_prefix_network')
    ccolor = weechat.color('chat')
    mp = weechat.info_get_hashtable(
        'irc_message_parse', {"message": string})
    buffer = weechat.buffer_search(
        "irc", "%s.%s" % (server, mp['channel']))
    if mp['tags']:
        tags = dict([s.split('=') for s in mp['tags'].split(';')])
        msg = tags['system-msg'].replace('\s',' ')
        if mp['text']:
            msg += ' [Comment] '+mp['text']
        weechat.prnt(buffer, '%s--%s %s' % (pcolor, ccolor, msg))
    return ''


def twitch_whisper(data, modifier, modifier_data, string):
    message = weechat.info_get_hashtable(
        'irc_message_parse', {"message": string})
    if message['tags']: string = '@'+message['tags']+' '
    else: string = ''
    string += ':'+message['host']
    string += ' PRIVMSG'
    string += ' '+message['arguments']
    return string


def twitch_privmsg(data, modifier, server_name, string):
    if not server_name in OPTIONS['servers'].split():
        return string
    message = weechat.info_get_hashtable(
        'irc_message_parse', {"message": string})
    if message['channel'].startswith('#'):
        return string
    newmsg = 'PRIVMSG #%s :/w %s %s' % (message['nick'],message['nick'],message['text'])
    return newmsg


def twitch_in_privmsg(data, modifier, server_name, string, prefix=''):
    if not OPTIONS['prefix_nicks']: return string
    if not server_name in OPTIONS['servers'].split():
        return string

    mp = weechat.info_get_hashtable("irc_message_parse", {"message": string})

    if not mp['tags']:
        return string
    if '#' + mp['nick'] == mp['channel']:
        return mp['message_without_tags'].replace(mp['nick'], '~' + mp['nick'], 1)

    tags = dict([s.split('=') for s in mp['tags'].split(';')])
    if tags['user-type'] == 'mod':
        prefix += '@'
    if tags['subscriber'] == '1':
        prefix += '%'
    if prefix:
        msg = mp['message_without_tags'].replace(
            mp['nick'], prefix + mp['nick'], 1)
        return '@' + mp['tags'] + ' ' + msg
    else:
        return string


def twitch_whois(data, modifier, server_name, string):
    if not server_name in OPTIONS['servers'].split():
        return string
    msg = weechat.info_get_hashtable("irc_message_parse", {"message": string})
    username = msg['nick'].lower()
    currentbuf = weechat.current_buffer()
    url = 'https://api.twitch.tv/kraken/users?login=' + username
    params='&api_version=5'
    url_hook = weechat.hook_process_hashtable(
        "url:" + url+params, curlopt, 7 * 1000, "channel_api", currentbuf)
    return ""

def config_setup():
    for option,value in OPTIONS.items():
        weechat.config_set_desc_plugin(option, '%s' % value[1])
        if not weechat.config_is_set_plugin(option):
            weechat.config_set_plugin(option, value[0])
            OPTIONS[option] = value[0]
        else:
            if option == 'prefix_nicks' or option == 'debug' or option == 'ssl_verify':
                OPTIONS[option] = weechat.config_string_to_boolean(
                    weechat.config_get_plugin(option))
                if option == 'debug':
                    if value == 0:
                        curlopt['verbose'] = "0"
                    else:
                        curlopt['verbose'] = "1"
                if option == 'ssl_verify':
                    if value == 0:
                        curlopt['ssl_verifypeer'] = "0"
                        curlopt['ssl_verifyhost'] = "0"
                    else:
                        curlopt['ssl_verifypeer'] = "1"
                        curlopt['ssl_verifyhost'] = "2"
            else:
                OPTIONS[option] = weechat.config_get_plugin(option)

def config_change(pointer, name, value):
    option = name.replace('plugins.var.python.'+SCRIPT_NAME+'.','')
    if option == 'prefix_nicks' or option == 'debug' or option == 'ssl_verify':
        value=weechat.config_string_to_boolean(value)
        if option == 'debug':
            if value == 0:
                curlopt['verbose'] = "0"
            if value == 1:
                curlopt['verbose'] = "1"
        if option == 'ssl_verify':
            if value == 0:
                curlopt['ssl_verifypeer'] = "0"
                curlopt['ssl_verifyhost'] = "0"
            if value == 1:
                curlopt['ssl_verifypeer'] = "1"
                curlopt['ssl_verifyhost'] = "2"
    OPTIONS[option] = value
    return weechat.WEECHAT_RC_OK


if weechat.register(SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION, SCRIPT_LICENSE,
                    SCRIPT_DESC, "", ""):
    weechat.hook_command("twitch", SCRIPT_DESC, "",
        "  settings:\n"
        "    plugins.var.python.twitch.servers (default: twitch)\n"
        "    plugins.var.python.twitch.prefix_nicks (default: 1)\n"
        "    plugins.var.python.twitch.debug (default: 0)\n"
        "    plugins.var.python.twitch.ssl_verify (default: 0)\n"
        "\n\n"
        "  This script checks stream status of any channel on any servers listed\n"
        "  in the \"plugins.var.python.twitch.servers\" setting. When you switch\n"
        "  to a buffer it will display updated infomation about the stream in the\n"
        "  title bar. Typing '/twitch' in a buffer will also fetch updated infomation.\n"
        "  '/whois nick' will lookup user info and display it in current buffer.\n\n"
        "  Option \"plugins.var.python.twitch.servers\" controls\n"
        "  what server this script will work on. The default is twitch\n"
        "  but you can have multiples separated by a space.\n"
        "  /set plugins.var.python.twitch.servers twitch twitchcopy\n"
        "\n\n"
        "  This script also will prefix users nicks (@ for mod, % for sub,\n"
        "  and ~ for broadcaster). This will break the traditional function\n"
        "  of `/ignore add nightbot` and will require you to prefix nicks if you\n"
        "  want to ignore someone `/ignore add re:[~@%]{0,3}nightbot` should ignore\n"
        "  a nick with all or none of the prefixes used by this script.\n"
        "  NOTE: This may cause high cpu usage in very active chat and/or on slower cpus.\n"
        "  This can also be disabled by setting\n    /set plugins.var.python.twitch.prefix_nicks off\n"
        "\n\n"
        "  If you are experiencing errors you can enable debug mode by setting\n"
        "    /set plugins.var.python.twitch.debug on\n"
        "  You can also try disabling SSL/TLS cert verification.\n"
        "    /set plugins.var.python.twitch.ssl_verify off\n"
        "\n\n"
        "  Required server settings:\n"
        "    /server add twitch irc.twitch.tv\n"
        "    /set irc.server.twitch.capabilities \"twitch.tv/membership,twitch.tv/commands,twitch.tv/tags\"\n"
        "    /set irc.server.twitch.nicks \"My Twitch Username\"\n"
        "    /set irc.server.twitch.password \"oauth:My Oauth Key\"\n"
        "\n"
        "  If you do not have a oauth token one can be generated for your account here\n"
        "    https://twitchapps.com/tmi/\n"
        "\n"
        "  This script also has whisper support that works like a standard query. \"/query user\"\n\n",
        "", "twitch_main", "")
    weechat.hook_signal('buffer_switch', 'twitch_buffer_switch', '')
    weechat.hook_config('plugins.var.python.' + SCRIPT_NAME + '.*', 'config_change', '')
    config_setup()
    weechat.hook_modifier("irc_in_CLEARCHAT", "twitch_clearchat", "")
    weechat.hook_modifier("irc_in_CLEARMSG", "twitch_clearmsg", "")
    weechat.hook_modifier("irc_in_RECONNECT", "twitch_reconnect", "")
    weechat.hook_modifier("irc_in_USERSTATE", "twitch_suppress", "")
    weechat.hook_modifier("irc_in_HOSTTARGET", "twitch_suppress", "")
    weechat.hook_modifier("irc_in_ROOMSTATE", "twitch_roomstate", "")
    weechat.hook_modifier("irc_in_USERNOTICE", "twitch_usernotice", "")
    weechat.hook_modifier("irc_in_WHISPER", "twitch_whisper", "")
    weechat.hook_modifier("irc_out_PRIVMSG", "twitch_privmsg", "")
    weechat.hook_modifier("irc_out_WHOIS", "twitch_whois", "")
    weechat.hook_modifier("irc_in_PRIVMSG", "twitch_in_privmsg", "")
